# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from jinja2.environment import Template
from uidom.dom.src.main import extension

__all__ = [
    "If",
    "Elif",
    "Else",
    "For",
    "AutoEscape",
    "Include",
    "Cycle",
    "Comment",
    "Extends",
    "Load",
    "CSRFToken",
    "Block",
    "Var",
    "render"
]


class JinjaConfig(object):
    self_dedent = False
    child_dedent = False
    enable_left_delimiter_space = True
    enable_right_delimiter_space = True


class JinjaDoubleTags(JinjaConfig, extension.DoubleTemplates):
    pass


class JinjaSingleTags(JinjaConfig, extension.SingleTemplates):
    pass


class Block(JinjaDoubleTags):

    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """
        super(Block, self).__init__("block", template_text, *dom_elements)


class For(JinjaDoubleTags):

    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """
        super(For, self).__init__("for", template_text, *dom_elements)


class If(JinjaDoubleTags):

    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """
        super(If, self).__init__("if", template_text, *dom_elements)


class Elif(JinjaSingleTags):
    self_dedent = True
    child_dedent = False

    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(Elif, self).__init__(
            "elif", template_text, *dom_elements,
        )


class Else(JinjaSingleTags):
    self_dedent = True
    child_dedent = False

    def __init__(self, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(Else, self).__init__("else", "", *dom_elements)


class AutoEscape(JinjaSingleTags):
    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(AutoEscape, self).__init__("autoescape", template_text, *dom_elements)


class Include(JinjaSingleTags):

    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(Include, self).__init__("include", template_text, *dom_elements)


class Cycle(JinjaSingleTags):
    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(Cycle, self).__init__("cycle", template_text, *dom_elements)


class Comment(JinjaDoubleTags):
    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(Comment, self).__init__("comment", template_text, *dom_elements)


class Extends(JinjaSingleTags):
    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(Extends, self).__init__("extends", template_text, *dom_elements)


class Load(JinjaSingleTags):
    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(Load, self).__init__("load", template_text, *dom_elements)


class CSRFToken(JinjaSingleTags):
    def __init__(self, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(CSRFToken, self).__init__("csrf_token", "", *dom_elements)


class Var(JinjaSingleTags):
    left_delimiter = "{{"
    right_delimiter = "}}"

    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(Var, self).__init__("", template_text, *dom_elements)


def render(template, **options):
    return Template(
        template.render() if isinstance(template, extension.Tags) else template,
        lstrip_blocks=True,
        trim_blocks=True,
        enable_async=True).render(**options)


if __name__ == '__main__':
    from collections import namedtuple as nt

    from uidom.dom.src.htmltags import a, html, li, nav, p, section, ul

    print(Block("base",
                html(For("name in names",
                         If("name", Block("load", Load("space")),
                            Elif("njnsf", p("ksf")),
                            Else(section(p("ok", Var("name"))))
                            )
                         )
                     ))
          )

    menu = Block(
        "nav",
        nav(
            ul(
                For(
                    "item in menu_items",
                    li(a(Var("item.name"), href=Var("item.link"))),
                )
            )
        ),
    )
    menu_url = nt("menu_url", "name link")
    print(render(
        menu,
        menu_items=[
            menu_url("Home", "home.html"),
            menu_url("About", "about.html"),
            menu_url("Contact Us", "contact_us.html")
        ]
    )
    )

    print(ul(For("name in names", li(Var("name")))))
    print(ul(li(x_text="name"), x_for="name in names", x_data={}))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class ObjectBottom(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-bottom"
        return elem


class ObjectCenter(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-center"
        return elem


class ObjectLeft(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-left"
        return elem


class ObjectLeftBottom(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-left-bottom"
        return elem


class ObjectLeftTop(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-left-top"
        return elem


class ObjectRight(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-right"
        return elem


class ObjectRightBottom(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-right-bottom"
        return elem


class ObjectRightTop(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-right-top"
        return elem


class ObjectTop(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "object-top"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.src.dom_tag import dom_tag
from uidom.dom.src.main import extension

try:
    basestring = basestring
except NameError:  # py3
    basestring = str
    unicode = str

underscored_classes = set(['del', 'input', 'map', 'object'])

# Tag attributes
_ATTR_GLOBAL = set([
    'accesskey', 'class', 'class', 'contenteditable', 'contextmenu', 'dir',
    'draggable', 'id', 'item', 'hidden', 'lang', 'itemprop', 'spellcheck',
    'style', 'subject', 'tabindex', 'title'
])
_ATTR_EVENTS = set([
    'onabort', 'onblur', 'oncanplay', 'oncanplaythrough', 'onchange', 'onclick',
    'oncontextmenu', 'ondblclick', 'ondrag', 'ondragend', 'ondragenter',
    'ondragleave', 'ondragover', 'ondragstart', 'ondrop', 'ondurationchange',
    'onemptied', 'onended', 'onerror', 'onfocus', 'onformchange', 'onforminput',
    'oninput', 'oninvalid', 'onkeydown', 'onkeypress', 'onkeyup', 'onload',
    'onloadeddata', 'onloadedmetadata', 'onloadstart', 'onmousedown',
    'onmousemove', 'onmouseout', 'onmouseover', 'onmouseup', 'onmousewheel',
    'onpause', 'onplay', 'onplaying', 'onprogress', 'onratechange',
    'onreadystatechange', 'onscroll', 'onseeked', 'onseeking', 'onselect',
    'onshow', 'onstalled', 'onsubmit', 'onsuspend', 'ontimeupdate',
    'onvolumechange', 'onwaiting'
])

ERR_ATTRIBUTE = 'attributes'
ERR_CONTEXT = 'context'
ERR_CONTENT = 'content'


class HtmlSingleTemplates(extension.SingleTemplates):
    self_dedent = False
    child_dedent = True
    enable_left_delimiter_space = False
    enable_right_delimiter_space = False


class HtmlDoubleTemplates(extension.DoubleTemplates):
    # self_dedent = True
    # child_dedent = True
    enable_left_delimiter_space = True
    enable_right_delimiter_space = True


class HtmlSingleTags(extension.SingleTags):
    pass


class HtmlDoubleTags(extension.DoubleTags):
    pass


class DocType(HtmlSingleTemplates):
    left_delimiter = "<!"
    right_delimiter = ">"

    def __init__(self, template_text="html", *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """

        super(DocType, self).__init__("DOCTYPE", template_text, *dom_elements)


class html_tag(extension.Tags):
    def __init__(self, *args, **kwargs):
        '''
        Creates a new html tag instance.
        '''
        super(html_tag, self).__init__(*args, **kwargs)

    # def validate(self):
    #   '''
    #   Validate the tag. This will check the attributes, context, and contents and
    #   emit tuples in the form of: element, message.
    #   '''
    #   errors = []

    #   errors.extend(self.validate_attributes())
    #   errors.extend(self.validate_context())
    #   errors.extend(self.validate_content())

    #   return errors

    # def validate_attributes(self):
    #   '''
    #   Validate the tag attributes.
    #   '''
    #   return []

    # def validate_context(self):
    #   '''
    #   Validate the tag context.
    #   '''
    #   return []

    # def validate_content(self):
    #   '''
    #   Validate the content of the tag.
    #   '''
    #   return []

    # def _check_attributes(self, *attrs):
    #   valid = set([])
    #   for attr in attrs:
    #     if hasattr(attr, '__iter__'):
    #       valid |= set(attr)
    #     else:
    #       valid.add(attr)
    #   return set(list(self.attributes.iterkeys())) - valid


################################################################################
############################### Html Tag Classes ###############################
################################################################################

# Root element

class html(html_tag):
    '''
    The html element represents the root of an HTML document.
    '''
    pass
    # def validate_attributes(self):
    #   errors = []
    #   for invalid in self._check_attributes(_ATTR_GLOBAL, 'manifest'):
    #     errors.append( (self, ERR_ATTRIBUTE, 'Invalid attribute: "%s"' % invalid) )
    #   return errors

    # def validate_context(self):
    #   if self.parent is not None and not isinstance(self.parent, iframe):
    #     return [(self, ERR_CONTEXT, 'Must be root element or child of an <iframe>')]
    #   return []

    # def validate_content(self):
    #   if len(self) != 2 or not isinstance(self[0], head) or not isinstance(self[1], body):
    #     return [(self, ERR_CONTENT, 'Children must be <head> and then <body>.')]
    #   return []


# Document metadata
class head(html_tag):
    '''
    The head element represents a collection of metadata for the document.
    '''
    pass


class title(html_tag):
    '''
    The title element represents the document's title or name. Authors should use
    titles that identify their documents even when they are used out of context,
    for example in a user's history or bookmarks, or in search results. The
    document's title is often different from its first heading, since the first
    heading does not have to stand alone when taken out of context.
    '''

    def _get_text(self):
        return u''.join(self.get(basestring))

    def _set_text(self, text):
        self.clear()
        self.add(text)

    text = property(_get_text, _set_text)


class base(html_tag):
    '''
    The base element allows authors to specify the document base URL for the
    purposes of resolving relative URLs, and the name of the default browsing
    context for the purposes of following hyperlinks. The element does not
    represent any content beyond this information.
    '''
    is_single = True


class link(html_tag):
    '''
    The link element allows authors to link their document to other resources.
    '''
    is_single = True


class meta(html_tag):
    '''
    The meta element represents various kinds of metadata that cannot be
    expressed using the title, base, link, style, and script elements.
    '''
    is_single = True


class style(html_tag):
    '''
    The style element allows authors to embed style information in their
    documents. The style element is one of several inputs to the styling
    processing model. The element does not represent content for the user.
    '''
    is_pretty = False


# Scripting
class script(html_tag):
    '''
    The script element allows authors to include dynamic script and data blocks
    in their documents. The element does not represent content for the user.
    '''
    is_pretty = True


class noscript(html_tag):
    '''
    The noscript element represents nothing if scripting is enabled, and
    represents its children if scripting is disabled. It is used to present
    different markup to user agents that support scripting and those that don't
    support scripting, by affecting how the document is parsed.
    '''
    pass


# Sections
class body(html_tag):
    '''
    The body element represents the main content of the document.
    '''
    pass


class main(html_tag):
    '''
    The main content area of a document includes content that is unique to that
    document and excludes content that is repeated across a set of documents such
    as site navigation links, copyright information, site logos and banners and
    search forms (unless the document or application's main function is that of a
    search form).
    '''


class section(html_tag):
    '''
    The section element represents a generic section of a document or
    application. A section, in this context, is a thematic grouping of content,
    typically with a heading.
    '''
    pass


class nav(html_tag):
    '''
    The nav element represents a section of a page that links to other pages or
    to parts within the page: a section with navigation links.
    '''
    pass


class article(html_tag):
    '''
    The article element represents a self-contained composition in a document,
    page, application, or site and that is, in principle, independently
    distributable or reusable, e.g. in syndication. This could be a forum post, a
    magazine or newspaper article, a blog entry, a user-submitted comment, an
    interactive widget or gadget, or any other independent item of content.
    '''
    pass


class aside(html_tag):
    '''
    The aside element represents a section of a page that consists of content
    that is tangentially related to the content around the aside element, and
    which could be considered separate from that content. Such sections are
    often represented as sidebars in printed typography.
    '''
    pass


class h1(html_tag):
    '''
    Represents the highest ranking heading.
    '''
    pass


class h2(html_tag):
    '''
    Represents the second-highest ranking heading.
    '''
    pass


class h3(html_tag):
    '''
    Represents the third-highest ranking heading.
    '''
    pass


class h4(html_tag):
    '''
    Represents the fourth-highest ranking heading.
    '''
    pass


class h5(html_tag):
    '''
    Represents the fifth-highest ranking heading.
    '''
    pass


class h6(html_tag):
    '''
    Represents the sixth-highest ranking heading.
    '''
    pass


class hgroup(html_tag):
    '''
    The hgroup element represents the heading of a section. The element is used
    to group a set of h1-h6 elements when the heading has multiple levels, such
    as subheadings, alternative titles, or taglines.
    '''
    pass


class header(html_tag):
    '''
    The header element represents a group of introductory or navigational aids.
    '''
    pass


class footer(html_tag):
    '''
    The footer element represents a footer for its nearest ancestor sectioning
    content or sectioning root element. A footer typically contains information
    about its section such as who wrote it, links to related documents,
    copyright data, and the like.
    '''
    pass


class address(html_tag):
    '''
    The address element represents the contact information for its nearest
    article or body element ancestor. If that is the body element, then the
    contact information applies to the document as a whole.
    '''
    pass


# Grouping content
class p(html_tag):
    '''
    The p element represents a paragraph.
    '''
    pass


class hr(html_tag):
    '''
    The hr element represents a paragraph-level thematic break, e.g. a scene
    change in a story, or a transition to another topic within a section of a
    reference book.
    '''
    is_single = True


class pre(html_tag):
    '''
    The pre element represents a block of preformatted text, in which structure
    is represented by typographic conventions rather than by elements.
    '''
    is_pretty = False


class blockquote(html_tag):
    '''
    The blockquote element represents a section that is quoted from another
    source.
    '''
    pass


class ol(html_tag):
    '''
    The ol element represents a list of items, where the items have been
    intentionally ordered, such that changing the order would change the
    meaning of the document.
    '''
    pass


class ul(html_tag):
    '''
    The ul element represents a list of items, where the order of the items is
    not important - that is, where changing the order would not materially change
    the meaning of the document.
    '''
    pass


class li(html_tag):
    '''
    The li element represents a list item. If its parent element is an ol, ul, or
    menu element, then the element is an item of the parent element's list, as
    defined for those elements. Otherwise, the list item has no defined
    list-related relationship to any other li element.
    '''
    pass


class dl(html_tag):
    '''
    The dl element represents an association list consisting of zero or more
    name-value groups (a description list). Each group must consist of one or
    more names (dt elements) followed by one or more values (dd elements).
    Within a single dl element, there should not be more than one dt element for
    each name.
    '''
    pass


class dt(html_tag):
    '''
    The dt element represents the term, or name, part of a term-description group
    in a description list (dl element).
    '''
    pass


class dd(html_tag):
    '''
    The dd element represents the description, definition, or value, part of a
    term-description group in a description list (dl element).
    '''
    pass


class figure(html_tag):
    '''
    The figure element represents some flow content, optionally with a caption,
    that is self-contained and is typically referenced as a single unit from the
    main flow of the document.
    '''
    pass


class figcaption(html_tag):
    '''
    The figcaption element represents a caption or legend for the rest of the
    contents of the figcaption element's parent figure element, if any.
    '''
    pass


class div(html_tag):
    '''
    The div element has no special meaning at all. It represents its children. It
    can be used with the class, lang, and title attributes to mark up semantics
    common to a group of consecutive elements.
    '''
    pass


# Text semantics
class a(html_tag):
    '''
    If the a element has an href attribute, then it represents a hyperlink (a
    hypertext anchor).

    If the a element has no href attribute, then the element represents a
    placeholder for where a link might otherwise have been placed, if it had been
    relevant.
    '''
    pass


class em(html_tag):
    '''
    The em element represents stress emphasis of its contents.
    '''
    pass


class strong(html_tag):
    '''
    The strong element represents strong importance for its contents.
    '''
    pass


class small(html_tag):
    '''
    The small element represents side comments such as small print.
    '''
    pass


class s(html_tag):
    '''
    The s element represents contents that are no longer accurate or no longer
    relevant.
    '''
    pass


class cite(html_tag):
    '''
    The cite element represents the title of a work (e.g. a book, a paper, an
    essay, a poem, a score, a song, a script, a film, a TV show, a game, a
    sculpture, a painting, a theatre production, a play, an opera, a musical, an
    exhibition, a legal case report, etc). This can be a work that is being
    quoted or referenced in detail (i.e. a citation), or it can just be a work
    that is mentioned in passing.
    '''
    pass


class q(html_tag):
    '''
    The q element represents some phrasing content quoted from another source.
    '''
    pass


class dfn(html_tag):
    '''
    The dfn element represents the defining instance of a term. The paragraph,
    description list group, or section that is the nearest ancestor of the dfn
    element must also contain the definition(s) for the term given by the dfn
    element.
    '''
    pass


class abbr(html_tag):
    '''
    The abbr element represents an abbreviation or acronym, optionally with its
    expansion. The title attribute may be used to provide an expansion of the
    abbreviation. The attribute, if specified, must contain an expansion of the
    abbreviation, and nothing else.
    '''
    pass


class time_(html_tag):
    '''
    The time element represents either a time on a 24 hour clock, or a precise
    date in the proleptic Gregorian calendar, optionally with a time and a
    time-zone offset.
    '''
    pass


_time = time_


class code(html_tag):
    '''
    The code element represents a fragment of computer code. This could be an XML
    element name, a filename, a computer program, or any other string that a
    computer would recognize.
    '''
    pass


class var(html_tag):
    '''
    The var element represents a variable. This could be an actual variable in a
    mathematical expression or programming context, an identifier representing a
    constant, a function parameter, or just be a term used as a placeholder in
    prose.
    '''
    pass


class samp(html_tag):
    '''
    The samp element represents (sample) output from a program or computing
    system.
    '''
    pass


class kbd(html_tag):
    '''
    The kbd element represents user input (typically keyboard input, although it
    may also be used to represent other input, such as voice commands).
    '''
    pass


class sub(html_tag):
    '''
    The sub element represents a subscript.
    '''
    pass


class sup(html_tag):
    '''
    The sup element represents a superscript.
    '''
    pass


class i(html_tag):
    '''
    The i element represents a span of text in an alternate voice or mood, or
    otherwise offset from the normal prose in a manner indicating a different
    quality of text, such as a taxonomic designation, a technical term, an
    idiomatic phrase from another language, a thought, or a ship name in Western
    texts.
    '''
    pass


class b(html_tag):
    '''
    The b element represents a span of text to which attention is being drawn for
    utilitarian purposes without conveying any extra importance and with no
    implication of an alternate voice or mood, such as key words in a document
    abstract, product names in a review, actionable words in interactive
    text-driven software, or an article lede.
    '''
    pass


class u(html_tag):
    '''
    The u element represents a span of text with an unarticulated, though
    explicitly rendered, non-textual annotation, such as labeling the text as
    being a proper name in Chinese text (a Chinese proper name mark), or
    labeling the text as being misspelt.
    '''
    pass


class mark(html_tag):
    '''
    The mark element represents a run of text in one document marked or
    highlighted for reference purposes, due to its relevance in another context.
    When used in a quotation or other block of text referred to from the prose,
    it indicates a highlight that was not originally present but which has been
    added to bring the reader's attention to a part of the text that might not
    have been considered important by the original author when the block was
    originally written, but which is now under previously unexpected scrutiny.
    When used in the main prose of a document, it indicates a part of the
    document that has been highlighted due to its likely relevance to the user's
    current activity.
    '''
    pass


class ruby(html_tag):
    '''
    The ruby element allows one or more spans of phrasing content to be marked
    with ruby annotations. Ruby annotations are short runs of text presented
    alongside base text, primarily used in East Asian typography as a guide for
    pronunciation or to include other annotations. In Japanese, this form of
    typography is also known as furigana.
    '''
    pass


class rt(html_tag):
    '''
    The rt element marks the ruby text component of a ruby annotation.
    '''
    pass


class rp(html_tag):
    '''
    The rp element can be used to provide parentheses around a ruby text
    component of a ruby annotation, to be shown by user agents that don't support
    ruby annotations.
    '''
    pass


class bdi(html_tag):
    '''
    The bdi element represents a span of text that is to be isolated from its
    surroundings for the purposes of bidirectional text formatting.
    '''
    pass


class bdo(html_tag):
    '''
    The bdo element represents explicit text directionality formatting control
    for its children. It allows authors to override the Unicode bidirectional
    algorithm by explicitly specifying a direction override.
    '''
    pass


class span(html_tag):
    '''
    The span element doesn't mean anything on its own, but can be useful when
    used together with the global attributes, e.g. class, lang, or dir. It
    represents its children.
    '''
    pass


class br(html_tag):
    '''
    The br element represents a line break.
    '''
    is_single = True
    is_inline = True


class wbr(html_tag):
    '''
    The wbr element represents a line break opportunity.
    '''
    is_single = True
    is_inline = True


# Edits
class ins(html_tag):
    '''
    The ins element represents an addition to the document.
    '''
    pass


class del_(html_tag):
    '''
    The del element represents a removal from the document.
    '''
    pass


_del = del_


# Embedded content
class img(html_tag):
    '''
    An img element represents an image.
    '''
    is_single = True


class iframe(html_tag):
    '''
    The iframe element represents a nested browsing context.
    '''
    pass


class embed(html_tag):
    '''
    The embed element represents an integration point for an external (typically
    non-HTML) application or interactive content.
    '''
    is_single = True


class object_(html_tag):
    '''
    The object element can represent an external resource, which, depending on
    the type of the resource, will either be treated as an image, as a nested
    browsing context, or as an external resource to be processed by a plugin.
    '''
    pass


_object = object_


class param(html_tag):
    '''
    The param element defines parameters for plugins invoked by object elements.
    It does not represent anything on its own.
    '''
    is_single = True


class video(html_tag):
    '''
    A video element is used for playing videos or movies, and audio files with
    captions.
    '''
    pass


class audio(html_tag):
    '''
    An audio element represents a sound or audio stream.
    '''
    pass


class source(html_tag):
    '''
    The source element allows authors to specify multiple alternative media
    resources for media elements. It does not represent anything on its own.
    '''
    is_single = True


class track(html_tag):
    '''
    The track element allows authors to specify explicit external timed text
    tracks for media elements. It does not represent anything on its own.
    '''
    is_single = True


class canvas(html_tag):
    '''
    The canvas element provides scripts with a resolution-dependent bitmap
    canvas, which can be used for rendering graphs, game graphics, or other
    visual images on the fly.
    '''
    pass


class map_(html_tag):
    '''
    The map element, in conjunction with any area element descendants, defines an
    image map. The element represents its children.
    '''
    pass


_map = map_


class area(html_tag):
    '''
    The area element represents either a hyperlink with some text and a
    corresponding area on an image map, or a dead area on an image map.
    '''
    is_single = True


# Tabular data
class table(html_tag):
    '''
    The table element represents data with more than one dimension, in the form
    of a table.
    '''
    pass


class caption(html_tag):
    '''
    The caption element represents the title of the table that is its parent, if
    it has a parent and that is a table element.
    '''
    pass


class colgroup(html_tag):
    '''
    The colgroup element represents a group of one or more columns in the table
    that is its parent, if it has a parent and that is a table element.
    '''
    pass


class col(html_tag):
    '''
    If a col element has a parent and that is a colgroup element that itself has
    a parent that is a table element, then the col element represents one or more
    columns in the column group represented by that colgroup.
    '''
    is_single = True


class tbody(html_tag):
    '''
    The tbody element represents a block of rows that consist of a body of data
    for the parent table element, if the tbody element has a parent and it is a
    table.
    '''
    pass


class thead(html_tag):
    '''
    The thead element represents the block of rows that consist of the column
    labels (headers) for the parent table element, if the thead element has a
    parent and it is a table.
    '''
    pass


class tfoot(html_tag):
    '''
    The tfoot element represents the block of rows that consist of the column
    summaries (footers) for the parent table element, if the tfoot element has a
    parent and it is a table.
    '''
    pass


class tr(html_tag):
    '''
    The tr element represents a row of cells in a table.
    '''
    pass


class td(html_tag):
    '''
    The td element represents a data cell in a table.
    '''
    pass


class th(html_tag):
    '''
    The th element represents a header cell in a table.
    '''
    pass


# Forms
class form(html_tag):
    '''
    The form element represents a collection of form-associated elements, some of
    which can represent editable values that can be submitted to a server for
    processing.
    '''
    pass


class fieldset(html_tag):
    '''
    The fieldset element represents a set of form controls optionally grouped
    under a common name.
    '''
    pass


class legend(html_tag):
    '''
    The legend element represents a caption for the rest of the contents of the
    legend element's parent fieldset element, if any.
    '''
    pass


class label(html_tag):
    '''
    The label represents a caption in a user interface. The caption can be
    associated with a specific form control, known as the label element's labeled
    control, either using for attribute, or by putting the form control inside
    the label element itself.
    '''
    pass


class input_(html_tag):
    '''
    The input element represents a typed data field, usually with a form control
    to allow the user to edit the data.
    '''
    is_single = True


_input = input_


class button(html_tag):
    '''
    The button element represents a button. If the element is not disabled, then
    the user agent should allow the user to activate the button.
    '''
    pass


class select(html_tag):
    '''
    The select element represents a control for selecting amongst a set of
    options.
    '''
    pass


class datalist(html_tag):
    '''
    The datalist element represents a set of option elements that represent
    predefined options for other controls. The contents of the element represents
    fallback content for legacy user agents, intermixed with option elements that
    represent the predefined options. In the rendering, the datalist element
    represents nothing and it, along with its children, should be hidden.
    '''
    pass


class optgroup(html_tag):
    '''
    The optgroup element represents a group of option elements with a common
    label.
    '''
    pass


class option(html_tag):
    '''
    The option element represents an option in a select element or as part of a
    list of suggestions in a datalist element.
    '''
    pass


class textarea(html_tag):
    '''
    The textarea element represents a multiline plain text edit control for the
    element's raw value. The contents of the control represent the control's
    default value.
    '''
    pass


class keygen(html_tag):
    '''
    The keygen element represents a key pair generator control. When the
    control's form is submitted, the private key is stored in the local keystore,
    and the public key is packaged and sent to the server.
    '''
    is_single = True


class output(html_tag):
    '''
    The output element represents the result of a calculation.
    '''
    pass


class progress(html_tag):
    '''
    The progress element represents the completion progress of a task. The
    progress is either indeterminate, indicating that progress is being made but
    that it is not clear how much more work remains to be done before the task is
    complete (e.g. because the task is waiting for a remote host to respond), or
    the progress is a number in the range zero to a maximum, giving the fraction
    of work that has so far been completed.
    '''
    pass


class meter(html_tag):
    '''
    The meter element represents a scalar measurement within a known range, or a
    fractional value; for example disk usage, the relevance of a query result, or
    the fraction of a voting population to have selected a particular candidate.
    '''
    pass


# Interactive elements
class details(html_tag):
    '''
    The details element represents a disclosure widget from which the user can
    obtain additional information or controls.
    '''
    pass


class summary(html_tag):
    '''
    The summary element represents a summary, caption, or legend for the rest of
    the contents of the summary element's parent details element, if any.
    '''
    pass


class command(html_tag):
    '''
    The command element represents a command that the user can invoke.
    '''
    is_single = True


class menu(html_tag):
    '''
    The menu element represents a list of commands.
    '''
    pass


class font(html_tag):
    '''
    The font element represents the font in a html .
    '''
    pass


# Additional markup
class comment(html_tag):
    is_inline = True
    '''
    Normal, one-line comment:
      >>> print comment("Hello, comments!")
      <!--Hello, comments!-->

    For IE's "if" statement comments:
      >>> print comment(p("Upgrade your browser."), condition='lt IE6')
      <!--[if lt IE6]><p>Upgrade your browser.</p><![endif]-->

    Downlevel conditional comments:
      >>> print comment(p("You are using a ", em("downlevel"), " browser."),
              condition='false', downlevel='revealed')
      <![if false]><p>You are using a <em>downlevel</em> browser.</p><![endif]>

    For more on conditional comments see:
      http://msdn.microsoft.com/en-us/library/ms537512(VS.85).aspx
    '''

    ATTRIBUTE_CONDITION = 'condition'

    # Valid values are 'hidden', 'downlevel' or 'revealed'
    ATTRIBUTE_DOWNLEVEL = 'downlevel'

    def _render(self, sb, indent_level=1, indent_str='  ', pretty=True, xhtml=False):
        has_condition = comment.ATTRIBUTE_CONDITION in self.attributes
        is_revealed = comment.ATTRIBUTE_DOWNLEVEL in self.attributes and \
                      self.attributes[comment.ATTRIBUTE_DOWNLEVEL] == 'revealed'

        sb.append('<!')
        if not is_revealed:
            sb.append('--')
        if has_condition:
            sb.append('[if %s]>' % self.attributes[comment.ATTRIBUTE_CONDITION])

        pretty = self._render_children(sb, indent_level+1, indent_str, pretty, xhtml)

        # if len(self.children) > 1:
        if any(isinstance(child, dom_tag) for child in self):
            sb.append('\n')
            sb.append(indent_str * indent_level)

        if has_condition:
            sb.append('<![endif]')
        if not is_revealed:
            if pretty:
                sb, _ = self._new_line_and_inline_handler(sb, indent_level, indent_str, pretty, self.is_inline)
            sb.append('--')
        sb.append('>')

        return sb


class template(html_tag):
    "The template element represents a template."
    pass


class slot(html_tag):
    "The slot element represents a slot."
    pass


class ConcatTag(extension.DoubleTags):
    # this is an empty tag with spits out children concatenated with new-line
    '''
    Supporting Concatenation of Tags as siblings with newline at end
    '''
    render_tag = False
    new_line_at_child_end = True


if __name__ == '__main__':
    from uidom.dom.src.utils.dom_util import text
    print(div(div(div(div(script("aa")), __inline=True))))
    print(ConcatTag(text('h')))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class Block(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " block"
        return elem


class InlineBlock(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " inline-block"
        return elem


class Inline(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " inline"
        return elem


class InlineFlex(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " inline-flex"
        return elem


class Table(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table"
        return elem


class InlineTable(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " inline-table"
        return elem


class TableCaption(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table-caption"
        return elem


class TableCell(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table-cell"
        return elem


class TableColumn(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table-column"
        return elem


class TableColumnGroup(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table-column-group"
        return elem


class TableHeaderGroup(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table-header-group"
        return elem


class TableFooterGroup(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table-header-group"
        return elem


class TableRowGroup(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table-row-group"
        return elem


class TableRow(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " table-row"
        return elem


class FlowRoot(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flow-root"
        return elem


class InlineGrid(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " inline-grid"
        return elem


class Contents(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " contents"
        return elem


class ListItem(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " list-item"
        return elem


class Hidden(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " hidden"
        return elem




# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.src import component


class VuetifyElement(component.Component):
    file_extension = ".vue"

    def __render__(self, *args, **kwargs):
        pass



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


__all__ = ["Element", "Sheet"]

VALUES_LINKED = list()


def link_values(list_a, list_b, values_linked: list = None or VALUES_LINKED):
    if not all(
            (isinstance(v, (tuple, list)) and len(v) == 2) for v in (list_a, list_b)
    ):
        raise TypeError(f"pass values as list or tuple")

    if not all(
            isinstance(v[0], Element) or isinstance(v[1], str) for v in (list_a, list_b)
    ):
        raise TypeError(
            f"not all values are of "
            f"[{Element.__name__}, {type(str).__name__}] types"
        )
    list_a = list(list_a)
    list_b = list(list_b)
    values_linked.append([list_a, list_b])


class Element(dict):
    values_linked = list()

    def __init__(self, **kwargs):
        super(Element, self).__init__(**kwargs)

    def link_values(self, attr, element, elem_attr):
        link_values([self, attr], [element, elem_attr], values_linked=self.values_linked)

    def __getattr__(self, item):
        return super(Element, self).__getitem__(item)

    def __missing__(self, key):
        return key

    def __setattr__(self, key, value):
        super(Element, self).__setitem__(key, value)
        for lst in self.values_linked:
            if [self, key] in lst:
                idx = lst.index([self, key])
                this = lst.pop(idx)
                other = lst.pop(0)
                other[0][other[1]] = value
                lst.extend([this, other])

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __str__(self):
        return str(dict(super(Element, self).items()))

    def __repr__(self):
        return f"{type(self).__qualname__}({self.__str__()})"

    def __dir__(self):
        return super().__dir__()


class FieldError(Exception):
    pass


class FieldToElementsMeta(type):
    def __new__(mcs, name, bases, namespace):  # NOQA
        _conflicting_fields = dict()
        for base in bases:
            if all(
                    [
                        any([base.__dict__.get("fields", None)]),
                        any([namespace.get("fields", None)]),
                    ]
            ):
                for f in namespace.get("fields"):
                    if f in base.__dict__.get("fields"):
                        if base not in _conflicting_fields:
                            _conflicting_fields[base] = list()
                        _conflicting_fields[base].append(f)

                for f in base.__dict__.get("fields"):
                    if not hasattr(base, f):
                        setattr(base, f, Element())
                    else:
                        if not isinstance(getattr(base, f), Element):
                            raise AttributeError(
                                f"{getattr(base, f)!r} must be instance of {Element!r}"
                            )

        if any(_conflicting_fields):
            raise FieldError(
                "conflicting fields found in {cf}".format(cf=_conflicting_fields)
            )
        return super(FieldToElementsMeta, mcs).__new__(mcs, name, bases, namespace)

    def __init_subclass__(mcs, **kwargs):
        if any([getattr(mcs, "fields", None)]):
            for f in getattr(mcs, "fields"):
                if not hasattr(mcs, f):
                    setattr(mcs, f, Element())
                else:
                    if not isinstance(getattr(mcs, f), Element):
                        raise AttributeError(
                            f"{getattr(mcs, f)!r} must be instance of {Element!r}"
                        )


class KeyToElementsAtrBase(object):
    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __setattr__(self, key, value):
        if key in type(self).__dict__ or key in self.__dict__:
            raise AttributeError(f"attribute [{key}] already exists")
        object.__setattr__(self, key, value)

    def __getattr__(self, item):
        if item not in type(self).__dict__:
            if item not in self.__dict__:
                object.__setattr__(self, item, Element())
            return self.__dict__[item]
        return type(self).__dict__[item]

    def __contains__(self, item):
        return hasattr(self, item)

    def __repr__(self):
        return repr(
            {i: f"{self.__dict__[i]!s}" for i in self.__dict__ if not i.startswith("_")}
        )


class Sheet(KeyToElementsAtrBase, metaclass=FieldToElementsMeta):
    def __init__(self, *args):
        self._args = args

    def __iter__(self):
        return iter(self._args)

    def __init_subclass__(cls, *args):
        super(Sheet, cls).__init__(cls, *args)

    def __dir__(self):
        return sorted(i for i in self.__dict__ if not i.startswith("_"))


class HeaderSheet(Sheet):
    fields = ["charset", "description", "lang", "link", "title", "viewport"]




# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from .clear import *
from .container import *
from .decoration import *
from .display import *
from .floats import *
from .isolation import *
from .objectfit import *
from .objectposition import *
from .overflow import *
from .overscroll import *
from .position import *
from .sizing import *
from .toprightbottomleft import *
from .visibility import *
from .zindex import *



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class ObjectContain(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " object-contain"
        return elem


class ObjectCover(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " object-cover"
        return elem


class ObjectFill(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " object-fill"
        return elem


class ObjectNone(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " object-none"
        return elem


class ObjectScaleDown(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " object-scale-down"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from dataclasses import dataclass, is_dataclass
#TODO: WIP
from datetime import datetime
from typing import get_args

from valio import (BOOL, CHOICE, DATE_TIME_DELTA, DEBUG, DEFAULT, DOC, INT,
                   NAME, PATTERN, STR, VALUE, TypeValidator, Validator,
                   email_pattern)


class EdgeQLValidator(Validator):
    exclusive: BOOL = TypeValidator(logger=False, debug=True)
    
    def __init__(self, 
                 default: DEFAULT = None, 
                 name: NAME = None, 
                 doc: DOC = None, 
                 required: BOOL = None, 
                 pattern: PATTERN = None, 
                 reassign: BOOL = None, 
                 multiple_of: VALUE = None, 
                 min_value: VALUE = None, 
                 value: VALUE = None, 
                 max_value: VALUE = None, 
                 gt: VALUE = None, 
                 eq: VALUE = None, 
                 lt: VALUE = None, 
                 min_length: INT = None, 
                 length: INT = None, 
                 max_length: INT = None, 
                 expire_after: DATE_TIME_DELTA = None, 
                 expire_on: DATE_TIME_DELTA = None, 
                 expire_before: DATE_TIME_DELTA = None, 
                 in_choice: CHOICE = None, 
                 not_in_choice: CHOICE = None, 
                 has_attributes: list[STR] = None, 
                 task_interval: INT = None, 
                 cache_task: BOOL = True, 
                 debug: DEBUG = None, 
                 cache_validation: BOOL = None, 
                 enable_async: BOOL = None, 
                 allow_validation: BOOL = True, 
                 exclusive: BOOL = None, 
                 **kwargs
                 ):
        
        self.exclusive = exclusive
        
        super().__init__(
            default=default,
            name=name,
            doc=doc,
            required=required,
            pattern=pattern,
            reassign=reassign,
            multiple_of=multiple_of,
            min_value=min_value,
            value=value,
            max_value=max_value,
            gt=gt,
            eq=eq,
            lt=lt,
            min_length=min_length,
            length=length,
            max_length=max_length,
            expire_after=expire_after,
            expire_on=expire_on,
            expire_before=expire_before,
            in_choice=in_choice,
            not_in_choice=not_in_choice,
            has_attributes=has_attributes,
            task_interval=task_interval,
            cache_task=cache_task,
            debug=debug,
            cache_validation=cache_validation,
            enable_async=enable_async,
            allow_validation=allow_validation,
            **kwargs)
    
    def __set_name__(self, owner, name):
        super(EdgeQLValidator, self).__set_name__(owner=owner, name=name)
        if getattr(owner, "create", None) is None:
            setattr(owner, "create", '')
            
        abstract = None 
        exclusive = None
        required = None 
        reassign = None 
        min_value = None
        max_value = None 
        min_length = None 
        max_length = None 
        pattern = None
        in_choice = None
        default = None
        constraint = None
        try:
            Meta = getattr(owner, 'Meta', None)
            abstract = getattr(Meta, 'abstract', None)
        except:
            pass 
        
        try:
            exclusive = self.exclusive
        except KeyError as ke:
            pass

        try:
            required = self.required
        except KeyError as ke: 
            pass 
        
        try: 
            reassign = self.reassign
        except KeyError as ke: 
            pass 
        
        try: 
            min_value = self.min_value
        except KeyError as ke: 
            pass 
        
        try: 
            max_value = self.max_value
        except KeyError as ke: 
            pass 
        
        try: 
            min_length = self.min_length
        except KeyError as ke: 
            pass
        
        try: 
            max_length = self.max_length
        except KeyError as ke: 
            pass
        
        try: 
            pattern = self.pattern
        except KeyError as ke: 
            pass
        
        try:
            in_choice = self.in_choice
        except KeyError as ke:
            pass 
        
        try:
            default = self.default
        except KeyError as ke:
            pass
        
        if any([exclusive, reassign is not None and not reassign, min_length, max_length, min_value, max_value, pattern, in_choice, default]):
            constraint = True 
        
        args = get_args(owner.__annotations__[name])
        if any(args):
            if any(get_args(args[0])):
                annotation_name = get_args(args[0])[0].__name__
            else:
                 annotation_name = args[0].__name__
        else:            
            annotation_name = self.annotation.__name__ if not isinstance(self.annotation, str) else self.annotation
             
        owner.create += f"{'abstract ' if abstract is not None and abstract else ''}type {owner.__name__}" + " {" if owner.create == '' else ''

        if owner.create.endswith("\n\t};\n}") or owner.create.endswith("\n}"):
            owner.create = owner.create[:-2]
        
        
        if not is_dataclass(owner.__annotations__[name] if not any(args) else args[0]):
            if required:
                owner.create += f"\n\trequired property {name} -> {annotation_name};" if not any(args) else f"\n\trequired multi property {name} -> array<<{annotation_name}>>;"
            else:
                owner.create += f"\n\tproperty {name} -> {annotation_name};" if not any(args) else f"\n\tmulti property {name} -> array<<{annotation_name}>>;"
        else:
            if required:
                owner.create += f"\n\trequired link {name} -> {annotation_name};" if not any(args) else f"\n\trequired multi link {name} -> {annotation_name};"
            else:
                owner.create += f"\n\tlink {name} -> {annotation_name};" if not any(args) else f"\n\tmulti link {name} -> {annotation_name};"

        if constraint:
            if owner.create.endswith(';'):
                owner.create = owner.create[:-1]
            owner.create += "\n\t{"

        owner.create += '''\n\t\tconstraint exclusive;''' if exclusive is not None and exclusive else ''
        owner.create += '''\n\t\treadonly := true;''' if not reassign and reassign is not None else ''
        owner.create += f'''\n\t\tconstraint min_value({min_value});''' if min_value and min_value is not None else ''
        owner.create += f'''\n\t\tconstraint max_value({max_value});''' if max_value and max_value is not None else ''
        owner.create += f'''\n\t\tconstraint min_len_value({min_length});''' if min_length and min_length is not None else ''
        owner.create += f'''\n\t\tconstraint max_len_value({max_length});''' if max_length and max_length is not None else ''
        owner.create += f'''\n\t\tconstraint regexp({pattern});''' if pattern and pattern is not None else ''
        owner.create += f'''\n\t\tconstraint one_of({in_choice});''' if in_choice and in_choice is not None else ''
        if not self.annotation is datetime:
            owner.create += f'''\n\t\tdefault := {self.default if not callable(self.default) else self.default.__qualname__};''' if default is not None else ''
        else:
            if self.default in [datetime.now, datetime.utcnow]:
                owner.create += f'''\n\t\tdefault := datetime_current()'''
                
        if constraint:
            owner.create += "\n\t}"
            
        owner.create += "\n}"

    
@dataclass
class User(object):
    create = ''
    created_at: datetime = EdgeQLValidator(logger=False, debug=True, required=True, default=datetime.utcnow)
    updated_at: datetime = EdgeQLValidator(logger=False, debug=True, reassign=True, default=datetime.utcnow)
    name: str = EdgeQLValidator(
        logger=False, 
        required=True, 
        reassign=False, 
        debug=True, 
        min_length=4,
        default='Guest',
        exclusive=True,
        )
    email_ids: list[str] | None = EdgeQLValidator(logger=False, debug=True, pattern=email_pattern)

    class Meta:
        abstract: bool = True

@dataclass
class SocialMedia(object):
    media_host: str = EdgeQLValidator(logger=False, debug=True, pattern=r"^https?://www.|www.\w{2,}.(net|com|edu|app|io|tech|store|biz|in|us|co.in)")
    media_handle: str = EdgeQLValidator(logger=False, debug=True)


@dataclass
class Account(object):
    users: list[User] = EdgeQLValidator(logger=False, debug=True, required=True, default=User(name="Guest"))
    media: list[SocialMedia] = EdgeQLValidator(logger=False, debug=True)
    
    
# if __name__ == '__main__':
    # print(Account(users=[User(name="User-1"), User(name="User-2")], 
    #               media=[SocialMedia(media_host="www.youtube.com", media_handle="unique1")]))
    # print(User.create)



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


__license__ = """
This file is part of Dominate.

Dominate is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

Dominate is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with Dominate.  If not, see
<http://www.gnu.org/licenses/>.
"""

# pylint: disable=bad-indentation, bad-whitespace, missing-docstring

import copy
import numbers
import threading
import typing
from collections import defaultdict, namedtuple
from functools import wraps

try:
    # Python 3
    from collections.abc import Callable
except ImportError:
    # Python 2.7
    from collections import Callable

try:
    basestring = basestring
except NameError:  # py3
    basestring = str
    unicode = str

try:
    import greenlet
except ImportError:
    greenlet = None


def _get_thread_context():
    context = [threading.current_thread()]
    if greenlet:
        context.append(greenlet.getcurrent())
    return hash(tuple(context))


class dom_tag(object):
    is_single = False  # Tag does not require matching end tag (ex. <hr/>)
    is_pretty = True  # Text inside the tag should be left as-is (ex. <pre>)
    # otherwise, text will be escaped() and whitespace may be
    # modified
    is_inline = False

    def __new__(_cls, *args, **kwargs):
        """
        Check if bare tag is being used as a decorator
        (called with a single function arg).
        decorate the function and return
        """
        if (
            len(args) == 1
            and isinstance(args[0], Callable)
            and not isinstance(args[0], dom_tag)
            and not kwargs
        ):
            wrapped = args[0]

            @wraps(wrapped)
            def f(*args, **kwargs):
                with _cls() as _tag:
                    return wrapped(*args, **kwargs) or _tag

            return f
        return object.__new__(_cls)

    def __init__(self, *args, **kwargs):
        """
        Creates a new tag. Child tags should be passed as arguments and attributes
        should be passed as keyword arguments.

        There is a non-rendering attribute which controls how the tag renders:

        * `__inline` - Boolean value. If True renders all children tags on the same
                       line.
        """

        self.attributes = {}
        self.children = []
        self.parent = None
        self.document = None

        # Does not insert newlines on all children if True (recursive attribute)
        self.is_inline = kwargs.pop("__inline", self.is_inline)
        self.is_pretty = kwargs.pop("__pretty", self.is_pretty)

        # Add child elements
        if args:
            self.add(*args)

        for attr, value in kwargs.items():
            self.set_attribute(*type(self).clean_pair(attr, value))

        self._ctx = None
        self._add_to_ctx()

    # context manager
    frame = namedtuple("frame", ["tag", "items", "used"])
    # stack of frames
    _with_contexts: typing.DefaultDict = defaultdict(list)

    def _add_to_ctx(self):
        stack = dom_tag._with_contexts.get(_get_thread_context())
        if stack:
            self._ctx = stack[-1]
            stack[-1].items.append(self)

    def __enter__(self):
        stack = dom_tag._with_contexts[_get_thread_context()]
        stack.append(dom_tag.frame(self, [], set()))
        return self

    def __exit__(self, type, value, traceback):
        thread_id = _get_thread_context()
        stack = dom_tag._with_contexts[thread_id]
        frame = stack.pop()
        for item in frame.items:
            if not hasattr(item, "__render__"):
                # deal with all the Tags but not Components
                if item in frame.used:
                    continue
            else:
                # deal with all the components here, as all components have ._entry attributes
                # we can now check the membership of it
                if item._entry in frame.used:
                    continue
            self.add(item)
        if not stack:
            del dom_tag._with_contexts[thread_id]

    def __call__(self, func):
        """
        tag instance is being used as a decorator.
        wrap func to make a copy of this tag
        """
        # remove decorator from its context so it doesn't
        # get added in where it was defined
        if self._ctx:
            self._ctx.used.add(self)

        @wraps(func)
        def f(*args, **kwargs):
            tag = copy.deepcopy(self)
            tag._add_to_ctx()
            with tag:
                return func(*args, **kwargs) or tag

        return f

    def set_attribute(self, key, value):
        """
        Add or update the value of an attribute.
        """
        if isinstance(key, int):
            self.children[key] = value
        elif isinstance(key, basestring):
            self.attributes[key] = value
        else:
            raise TypeError(
                "Only integer and string types are valid for assigning "
                "child tags and attributes, respectively."
            )

    __setitem__ = set_attribute

    def delete_attribute(self, key):
        if isinstance(key, int):
            del self.children[key : key + 1]
        else:
            del self.attributes[key]

    __delitem__ = delete_attribute

    def setdocument(self, doc):
        """
        Creates a reference to the parent document to allow for partial-tree
        validation.
        """
        # assume that a document is correct in the subtree
        if self.document != doc:
            self.document = doc
            for i in self.children:
                if not isinstance(i, dom_tag):
                    return
                i.setdocument(doc)

    def add(self, *args):
        """
        Add new child tags.
        """
        for obj in args:
            if isinstance(obj, numbers.Number):
                # Convert to string so we fall into next if block
                obj = str(obj)

            if isinstance(obj, basestring):
                obj = escape(obj)
                self.children.append(obj)

            elif isinstance(obj, dom_tag):
                stack = dom_tag._with_contexts.get(_get_thread_context())
                if stack:
                    if hasattr(obj, 'render_tag'):
                        if obj.render_tag:
                            stack[-1].used.add(obj)
                        else:
                            stack[-1].used.add(obj.children[0])
                    else:
                        stack[-1].used.add(obj)
                        
                self.children.append(obj)
                obj.parent = self
                obj.setdocument(self.document)

            elif isinstance(obj, dict):
                for attr, value in obj.items():
                    self.set_attribute(*dom_tag.clean_pair(attr, value))

            elif hasattr(obj, "__iter__"):
                for subobj in obj:
                    self.add(subobj)

            else:  # wtf is it?
                raise ValueError(self.__class__, "%r not a tag or string." % obj)

        if len(args) == 1:
            return args[0]

        return args

    def add_raw_string(self, s):
        self.children.append(s)

    def remove(self, obj):
        self.children.remove(obj)

    def clear(self):
        for i in self.children:
            if isinstance(i, dom_tag) and i.parent is self:
                i.parent = None
        self.children = []

    def get(self, tag=None, **kwargs):
        """
        Recursively searches children for tags of a certain
        type with matching attributes.
        """
        # Stupid workaround since we can not use dom_tag in the method declaration
        if tag is None:
            tag = dom_tag

        attrs = [
            (self.clean_attribute(attr), value) for attr, value in kwargs.items()
        ]

        results = []
        for child in self.children:
            if (isinstance(tag, basestring) and type(child).__name__ == tag) or (
                not isinstance(tag, basestring) and isinstance(child, tag)
            ):

                if all(
                    child.attributes.get(attribute) == value
                    for attribute, value in attrs
                ):
                    # If the child is of correct type and has all attributes and values
                    # in kwargs add as a result
                    results.append(child)
            if isinstance(child, dom_tag):
                # If the child is a dom_tag extend the search down through its children
                results.extend(child.get(tag, **kwargs))
        return results

    def __getitem__(self, key):
        """
        Returns the stored value of the specified attribute or child
        (if it exists).
        """
        if isinstance(key, int):
            # Children are accessed using integers
            try:
                return object.__getattribute__(self, "children")[key]
            except KeyError:
                raise IndexError('Child with index "%s" does not exist.' % key)
        elif isinstance(key, basestring):
            # Attributes are accessed using strings
            try:
                return object.__getattribute__(self, "attributes")[key]
            except KeyError:
                raise AttributeError('Attribute "%s" does not exist.' % key)
        else:
            raise TypeError(
                "Only integer and string types are valid for accessing "
                "child tags and attributes, respectively."
            )

    __getattr__ = __getitem__

    def __len__(self):
        """
        Number of child elements.
        """
        return len(self.children)

    def __bool__(self):
        """
        Hack for "if x" and __len__
        """
        return True

    __nonzero__ = __bool__

    def __iter__(self):
        """
        Iterates over child elements.
        """
        return self.children.__iter__()

    def __contains__(self, item):
        """
        Checks recursively if item is in children tree.
        Accepts both a string and a class.
        """
        return bool(self.get(item))

    def __iadd__(self, obj):
        """
        Reflexive binary addition simply adds tag as a child.
        """
        self.add(obj)
        return self

    # String and unicode representations are the same as render()
    def __unicode__(self):
        return self.render()

    __str__ = __unicode__

    def render(self, indent="  ", pretty=True, xhtml=False):
        data = self._render([], 0, indent, pretty, xhtml)
        return u"".join(data)

    def _render(self, sb, indent_level, indent_str, pretty, xhtml):
        pretty = pretty and self.is_pretty

        name = getattr(self, "tagname", type(self).__name__)

        # Workaround for python keywords and standard classes/methods
        # (del, object, input)
        if name[-1] == "_":
            name = name[:-1]

        # open tag
        sb.append("<")
        sb.append(name)

        for attribute, value in sorted(self.attributes.items()):
            if value is not False:  # False values must be omitted completely
                sb.append(' %s="%s"' % (attribute, escape(unicode(value), True)))

        sb.append(" />" if self.is_single and xhtml else ">")

        if not self.is_single:
            inline = self._render_children(
                sb, indent_level + 1, indent_str, pretty, xhtml
            )

            if pretty and not inline:
                sb.append("\n")
                sb.append(indent_str * indent_level)

            # close tag
            sb.append("</")
            sb.append(name)
            sb.append(">")

        return sb

    def _render_children(self, sb, indent_level, indent_str, pretty, xhtml):
        inline = True
        for child in self.children:
            if isinstance(child, dom_tag):
                if pretty and not child.is_inline:
                    inline = False
                    sb.append("\n")
                    sb.append(indent_str * indent_level)
                child._render(sb, indent_level, indent_str, pretty, xhtml)
            else:
                sb.append(unicode(child))

        return inline

    def __repr__(self):
        name = "%s.%s" % (self.__module__, type(self).__name__)

        attributes_len = len(self.attributes)
        attributes = "%s attribute" % attributes_len
        if attributes_len != 1:
            attributes += "s"

        children_len = len(self.children)
        children = "%s child" % children_len
        if children_len != 1:
            children += "ren"

        return "<%s at %x: %s, %s>" % (name, id(self), attributes, children)

    @staticmethod
    def clean_attribute(attribute):
        """
        Normalize attribute names for shorthand and work arounds for limitations
        in Python's syntax
        """

        # Shorthand
        attribute = {
            "cls": "class",
            "className": "class",
            "class_name": "class",
            "fr": "for",
            "html_for": "for",
            "htmlFor": "for",
        }.get(attribute, attribute)

        # Workaround for Python's reserved words
        if attribute[0] == "_":
            attribute = attribute[1:]

        # Workaround for dash
        special_prefix = any([attribute.startswith(x) for x in ("data_", "aria_")])
        if attribute in {"http_equiv"} or special_prefix:
            attribute = attribute.replace("_", "-")

        # Workaround for colon
        if attribute.split("_")[0] in ("xlink", "xml", "xmlns"):
            attribute = attribute.replace("_", ":", 1)

        return attribute

    @classmethod
    def clean_pair(cls, attribute, value):
        """
        This will call `clean_attribute` on the attribute and also allows for the
        creation of boolean attributes.

        Ex. input(selected=True) is equivalent to input(selected="selected")
        """
        attribute = cls.clean_attribute(attribute)

        # Check for boolean attributes
        # (i.e. selected=True becomes selected="selected")
        if value is True:
            value = attribute

        # Ignore `if value is False`: this is filtered out in render()

        return (attribute, value)


_get_current_none = object()


def get_current(default=_get_current_none):
    """
    get the current tag being used as a with context or decorated function.
    if no context is active, raises ValueError, or returns the default, if provided
    """
    h = _get_thread_context()
    ctx = dom_tag._with_contexts.get(h, None)
    if ctx:
        return ctx[-1].tag
    if default is _get_current_none:
        raise ValueError("no current context")
    return default


def attr(*args, **kwargs):
    """
    Set attributes on the current active tag context
    """
    c = get_current()
    dicts = args + (kwargs,)
    for d in dicts:
        for attr, value in d.items():
            c.set_attribute(*dom_tag.clean_pair(attr, value))


# escape() is used in render
from uidom.dom.src.utils.dom_util import escape



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from .component import *
from .csstags import *
from .ext import *
from .htmltags import *
from .jinjatags import *
from .svgtags import *
from .utils import *
from .xmltags import *



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from functools import wraps

# Taken from https://docs.datasette.io/en/stable/plugin_hooks.html#asgi-wrapper-datasette


def asgi_wrapper(dom):
    def wrap_with_dom_header(app):
        @wraps(app)
        async def add_x_dom_header(scope, receive, send):
            async def wrapped_send(event):
                if event["type"] == "http.response.start":
                    original_headers = event.get("headers") or []
                    event = {
                        "type": event["type"],
                        "status": event["status"],
                        "headers": original_headers + [
                            [b"x-dom",
                            ", ".join(dom.attributes.keys()).encode("utf-8")]
                        ],
                    }
                await send(event)
            await app(scope, receive, wrapped_send)
        return add_x_dom_header
    return wrap_with_dom_header



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import asyncio
import typing
import uuid
from dataclasses import dataclass
from datetime import datetime, time
from enum import Enum, IntEnum
from pathlib import Path
from uuid import UUID

import valio
from tortoise import Model, Tortoise
from uidom import dom

valio.Validator.register(Model)


class ModelValidator(valio.Validator):
    annotation = typing.Union[Model, None]


class ModelField(valio.Field):
    validator = ModelValidator


@dataclass
class Queries(object):
    model_field = ModelField(logger=True, debug=True)
    model: typing.Type[Model] = model_field.validator

    async def all(self):
        return await self.model.all()

    async def only(self, *field_names):
        return await self.model.all().only(*field_names)

    async def get(self, field: str, value=None):
        field = {f"{field}": value}
        value = await self.model.filter(**field).get()
        return value

    async def create(self, **kwargs):
        return await self.model.create(**kwargs)

    async def update(self, field: str, value=None, **kwargs):
        field = {f"{field}": value}
        value = await self.model.filter(**field).update(**kwargs)
        return value

    async def delete(self, field: str, value=None):
        field = {f"{field}": value}
        await self.model.filter(**field).delete()

    async def gt(self, field: str, value=None):
        field = f"{field}__gt"
        return await self.get(field=field, value=value)

    async def lt(self, field: str, value=None):
        field = f"{field}__lt"
        return await self.get(field=field, value=value)

    async def contains(self, field: str, value=None):
        field = f"{field}__contains"
        return await self.get(field=field, value=value)


async def init():
    # Here we create a SQLite DB using file "db.sqlite3"
    #  also specify the app name of "models"
    #  which contain models from "app.models"
    await Tortoise.init(
        db_url='sqlite://db.sqlite3',
        modules={'models': ['__main__']}
    )
    # Generate the schema
    await Tortoise.generate_schemas()


async def query(q):
    await init()
    value = await asyncio.gather(q)
    await Tortoise.close_connections()
    return value


class UiMeta(type):
    labels = None
    inputs = None
    fields = None
    buttons = None
    form = None

    INPUT = {
        str: dom.CharInput,
        int: dom.IntegerField,
        float: dom.FloatInput,
        bytes: dom.FileButtonInput,
        bool: dom.CheckboxInput,
        time: dom.DateInput,
        Path: dom.FileButtonInput,
        UUID: dom.CharInput,
        Enum: dom.EnumInput,
        IntEnum: dom.IntegerEnumInput
    }

    LABEL = {
        str: dom.CharLabel,
        int: dom.IntegerLabel,
        float: dom.FloatLabel,
        bytes: dom.CharLabel,
        bool: dom.BooleanLabel,
        datetime.time: dom.DateLabel,
        Path: dom.CharLabel,
        UUID: dom.CharLabel,
        Enum: dom.EnumLabel,
        IntEnum: dom.EnumLabel
    }

    CSS = {
        str: {
            "labels": "font-sm font-mono text-md px-2 uppercase",
            "inputs": "px-2 focus:outline-none focus:ring-2 focus:ring-gray-300 rounded-lg "
                      "placeholder-gray-400 bg-gray-100",
            "fields": "flex flex-row justify-between space-x-4 m-1 p-2 " 
                      "rounded-lg shadow-md border-2 border-gray-300 w-full"
        },

        bool: {
            "labels": "font-sm font-mono text-md px-2 uppercase",
            "inputs": "px-2 focus:outline-none focus:ring-2 focus:ring-gray-300 rounded-lg bg-gray-100",
            "fields": "flex flex-row justify-between space-x-2 m-1 p-2 "
                      "rounded-lg shadow-md border-2 border-gray-300 w-full"
        },

        "buttons": {"save": "flex flex-row bg-green-500 hover:bg-green-400 whitespace-nowrap "
                            "text-white m-2 p-2 rounded-lg justify-evenly space-x-2",
                    "delete": "flex flex-row border border-red-500 hover:border-red-400 whitespace-nowrap "
                              "text-red-500 m-2 p-2 rounded-lg justify-evenly space-x-2",
                    "edit": "flex flex-row bg-blue-500 hover:bg-blue-400 whitespace-nowrap "
                            "m-2 text-white p-2 rounded-lg justify-evenly space-x-2",
                    "buttons": "flex flex-row p-2 m-2 justify-between"}

        }

    def __new__(mcs, name, bases, namespaces):
        inputs = dict()
        labels = dict()
        fields = dict()
        _annotations = dict()
        buttons = dict()
        btn_css = mcs.CSS.get("buttons", None)
        meta = namespaces.get('Meta', None)
        model = getattr(meta, 'model', None)

        for base in bases:
            for ns, ann in base.__dict__.get("__annotations__", {}).items():
                if ns in getattr(base, "Meta", mcs.Meta).exclude:
                    continue
                _annotations[ns] = ann
                if not issubclass(ann, UIBase):
                    inputs[ns] = mcs.INPUT[ann]
                    labels[ns] = mcs.LABEL[ann]
                else:
                    data = queries(model.all().select_related(ns))
                    labels[ns] = dom.span(f"Choose {ns}")
                    inputs[ns] = dom.select(dom.option(opt, value=str(opt)) for opt in data)
                    edit_btn = dom.SubmitButton(label=f"Edit {ann.__qualname__}",
                                                     value=f"edit_{ann.__qualname__.lower()}")
                    edit_btn["class"] = btn_css.get("edit", '') if btn_css is not None else ''
                    fields[ns] = dom.div(labels[ns], inputs[ns], edit_btn)

        for ns, ann in namespaces.get("__annotations__", {}).items():
            if ns in getattr(namespaces, "Meta", mcs.Meta).exclude:
                continue
            _annotations[ns] = ann
            if not issubclass(ann, UIBase):
                inputs[ns] = mcs.INPUT[ann]
                labels[ns] = mcs.LABEL[ann]
            else:
                data = asyncio.run(query(model.all().select_related(ns)))
                labels[ns] = dom.span(f"Choose {ns}")
                inputs[ns] = dom.select(dom.option(id=opt.id, value=str(opt)) for opt in data)
                edit_btn = dom.SubmitButton(label=f"Edit {ann.__qualname__}", icon=dom.manage_icon,
                                                 value=f"edit_{ann.__qualname__.lower()}")
                edit_btn["class"] = btn_css.get("edit", '') if btn_css is not None else ''
                fields[ns] = dom.div(labels[ns], inputs[ns], edit_btn)

        for atr in inputs:
            if not isinstance(inputs[atr], dict):
                if callable(inputs[atr]):
                    atr_name = atr.replace("_", "-")
                    if "_" in atr:
                        atr_label = " ".join(map(lambda x: x.capitalize(), atr.replace('_', ' ').split()))
                    else:
                        atr_label = atr.capitalize()
                    css = mcs.CSS.get(_annotations[atr], None)
                    inputs[atr] = inputs[atr](name=atr_name, placeholder=f"Enter {atr_label}",
                                              cls=css.get("inputs", '') if css is not None else '')
                    labels[atr] = labels[atr](label=atr_label,
                                              cls=css.get("labels", '') if css is not None else '')
                    labels[atr]["for"] = inputs[atr]["id"] = atr_name
                    fields[atr] = dom.div(labels[atr], inputs[atr],
                                               cls=css.get("fields", '') if css is not None else '')

        buttons["save"] = dom.SubmitButton(label=f"Save {name}", icon=dom.save_icon,
                                                value=f'save_{name.lower()}')
        buttons["save"]["class"] = btn_css.get("save", '') if btn_css is not None else ''
        buttons["delete"] = dom.SubmitButton(label=f"Delete {name}", icon=dom.delete_icon,
                                                  value=f'delete_{name.lower()}')
        buttons["delete"]["class"] = btn_css.get("delete", '') if btn_css is not None else ''
        buttons["buttons"] = dom.div(buttons["delete"], buttons["save"],
                                          cls=btn_css.get("buttons", '') if btn_css is not None else '')

        form = dom.Form(*fields.values(), buttons["buttons"])
        cls = type.__new__(mcs, name, bases, namespaces)
        setattr(cls, "labels", labels)
        setattr(cls, "inputs", inputs)
        setattr(cls, "fields", fields)
        setattr(cls, "buttons", buttons)
        setattr(cls, "form", form)
        return cls

    class Meta:
        exclude:list = []


@dataclass
class UIBase(metaclass=UiMeta):
    pass


if __name__ == '__main__':
    import typing
    from dataclasses import dataclass

    import valio
    from tortoise import Model, fields
    from uidom.backend.database.db import queries

    class IDMixin(object):
        id = fields.UUIDField(pk=True)

    class TimeStampMixin(object):
        created_at = fields.DatetimeField(auto_now_add=True)
        updated_at = fields.DatetimeField(auto_now=True)


    class CountryModel(Model, IDMixin, TimeStampMixin):
        country = fields.CharField(max_length=50)

        def __str__(self):
            return f"Country: {str(self.country)}"

        class Meta:
            table = "country"

    class StateModel(Model, IDMixin, TimeStampMixin):
        country = fields.ForeignKeyField(model_name="models.CountryModel", related_name="states")
        state = fields.CharField(max_length=50)

        def __str__(self):
            return f"State: {str(self.state)}, {str(self.country)}"

        class Meta:
            table = "state"

    class CityModel(Model, IDMixin, TimeStampMixin):
        state = fields.ForeignKeyField(model_name="models.StateModel", related_name="cities")
        city = fields.CharField(max_length=50)

        def __str__(self):
            return f"City: {self.city}, {str(self.state)}"

        class Meta:
            table = "city"

    class LocationModel(Model, IDMixin, TimeStampMixin):
        city = fields.ForeignKeyField(model_name="models.CityModel", related_name="locations")
        street_name = fields.CharField(max_length=50)

        class Meta:
            table = "location"

    class UserModel(Model, IDMixin, TimeStampMixin):
        first_name = fields.CharField(max_length=50, blank=True, null=True)
        last_name = fields.CharField(max_length=50, blank=True, null=True)
        photo = fields.BinaryField()

        class Meta:
            table = "user"

    class AccountTypeEnum(str, Enum):
        MANUFACTURER = "manufacturer"
        RETAILER = "retailer"
        WHOLESALER = "wholesaler"
        BROKER = "broker"

    class ProductTypeEnum(str, Enum):
        DIAMOND = "diamond"
        BULLION = "bullion"
        JEWELLERY = "jewellery"
        COLOR_STONE = "color_stone"

    class AccountTypeModel(Model, IDMixin, TimeStampMixin):
        account_type = fields.CharEnumField(enum_type=AccountTypeEnum, unique=True)

        class Meta:
            table = "account_type"

    class ProductTypeModel(Model, IDMixin, TimeStampMixin):
        product_type = fields.CharEnumField(enum_type=ProductTypeEnum, unique=True)

        class Meta:
            table = "product_type"

    class AccountModel(Model, IDMixin, TimeStampMixin):
        email = fields.CharField(max_length=50)
        password = fields.CharField(max_length=80)
        is_active = fields.BooleanField(default=False)
        is_verified = fields.BooleanField(default=False)
        user = fields.ForeignKeyField(model_name="models.UserModel", related_name="accounts")
        product_type = fields.ForeignKeyField(model_name="models.ProductTypeModel", related_name="accounts")
        account_type = fields.ForeignKeyField(model_name="models.AccountTypeModel", related_name="accounts")
        location = fields.ForeignKeyField(model_name="models.LocationModel", related_name="accounts")

        class Meta:
            table = "account"

    class Tokens(Model, IDMixin, TimeStampMixin):
        account = fields.ForeignKeyField(model_name="models.Account")

        class Meta:
            table = "token"

    @dataclass
    class ID(object):
        id = valio.UUIDField(default=uuid.uuid4())

        class Meta:
            exclude = ["id"]

    @dataclass
    class TimeStamp(object):
        created_at: time = valio.DateValidator(logger=False, debug=True, default=time())
        updated_at: time = valio.DateValidator(logger=False, debug=True, default=time())

        class Meta:
            exclude = ["created_at", "updated_at"]

    @dataclass
    class Country(UIBase, ID, TimeStamp):
        country_field = valio.StringField(logger=False, max_length=50)
        country: str = country_field.validator

        class Meta:
            model = CountryModel

    valio.Validator.register(Country)  # noqa


    class CountryValidator(valio.Validator):
        annotation = typing.Union[Country, None]


    class CountryField(valio.Field):
        validator = CountryValidator


    @dataclass
    class State(UIBase, ID, TimeStamp):
        country_field = CountryField(logger=False, debug=True)
        state_field = valio.StringField(logger=False, max_length=50)

        country: typing.Union[Country, CountryValidator] = country_field.validator
        state: str = state_field.validator

        class Meta:
            model = StateModel

    valio.Validator.register(State)  # noqa

    class StateValidator(valio.Validator):
        annotation = typing.Union[State, None]

    class StateField(valio.Field):
        validator = StateValidator

    @dataclass
    class City(UIBase, ID, TimeStamp):
        state_field = StateField(logger=False)
        city_field = valio.StringField(logger=False, max_length=50)
        state: State = state_field.validator
        city: str = city_field.validator

        class Meta:
            model = CityModel

    valio.Validator.register(City)  # noqa

    class CityValidator(valio.Validator):
        annotation = typing.Union[City, None]

    class CityField(valio.Field):
        validator = CityValidator

    @dataclass
    class Location(UIBase, ID,  TimeStamp):
        city_field = CityField(logger=False)
        street_field = valio.StringField(logger=False)

        city: City = city_field.validator
        street_name = street_field.validator

        class Meta:
            model = LocationModel

    valio.Validator.register(Location)  # noqa

    class LocationValidator(valio.Validator):
        annotation = typing.Union[Location, None]

    @dataclass
    class User(UIBase, ID, TimeStamp):
        first_name: str = valio.StringValidator(logger=False, debug=True)
        last_name: str = valio.StringValidator(logger=False, debug=True)
        photo: Path = valio.PathValidator(logger=False, debug=True)

        class Meta:
            model = UserModel

    valio.Validator.register(User)  # noqa

    class UserValidator(valio.Validator):
        annotation = typing.Union[User, None]

    @dataclass
    class Account(UIBase, TimeStamp):
        email: str = valio.EmailIDValidator(logger=False, debug=True)
        is_active: bool = valio.BooleanValidator(logger=False, debug=True)
        is_verified: bool = valio.BooleanValidator(logger=False, debug=True)
        user: User = UserValidator(logger=False, debug=True)
        account_type: typing.Union[str, Enum, None] = valio.StringEnumField(in_choices=AccountTypeEnum, debug=True)
        product_type: typing.Union[str, Enum, None] = valio.StringEnumField(in_choices=ProductTypeEnum, debug=True)
        locations: Location = LocationValidator(logger=False, debug=True)

        class Meta:
            model = AccountModel

    print(Country.form)

    # user_form = Application(Country.form)
    # user_form.link_styles["href"] = "../production_path/static/file/css/styles.css"
    # user_form.save()



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement
from valio import Validator


class Grid(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " grid"
        return elem


class Columns(HTMLElement):

    def __render__(self, elem, columns):
        elem["class"] += f" grid-cols-{columns}"
        return elem


class ColumnAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " col-auto"
        return elem


class ColumnStart(HTMLElement):

    def __render__(self, elem, start="auto"):
        elem["class"] += f" col-start-{start}"
        return elem


class ColumnSpan(HTMLElement):

    def __render__(self, elem, span="full"):
        elem["class"] += f" col-span-{span}"
        return elem


class ColumnEnd(HTMLElement):

    def __render__(self, elem, end="auto"):
        elem["class"] += f" col-end-{end}"
        return elem


class Rows(HTMLElement):

    def __render__(self, elem, rows="none"):
        elem["class"] += f" grid-rows-{rows}"
        return elem


class RowAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += f" row-auto"
        return elem


class RowStart(HTMLElement):

    def __render__(self, elem, start="auto"):
        elem["class"] += f" row-start-{start}"
        return elem


class RowSpan(HTMLElement):

    def __render__(self, elem, span="full"):
        elem["class"] += f" row-span-{span}"
        return elem


class RowEnd(HTMLElement):

    def __render__(self, elem, end="auto"):
        elem["class"] += f" row-span-{end}"
        return elem


class GridFlow(HTMLElement):
    flow: str = Validator(logger=False, debug=True, in_choice=["row", "col", "row-dense", "col-dense"])

    def __render__(self, elem, flow="row"):
        self.flow = flow
        elem["class"] += f" grid-flow-{flow}"
        return elem


class GridAutoColumn(HTMLElement):
    auto_col: str = Validator(logger=False, debug=True, in_choice=["auto", "min", "max", "fr"])

    def __render__(self, elem, auto_col="auto"):
        self.auto_col = auto_col
        elem["class"] += f" auto-cols-{auto_col}"
        return elem


class GridAutoRow(HTMLElement):
    auto_row: str = Validator(logger=False, debug=True, in_choice=["auto", "min", "max", "fr"])

    def __render__(self, elem, auto_row="auto"):
        self.auto_row = auto_row
        elem["class"] += f" auto-cols-{auto_row}"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from collections import namedtuple as nt

# import pytest
from uidom.dom.src import Block, For, Var, a, li, nav, render, ul


def block_test():
    menu_url = nt("menu_url", "name link")
    menu = Block(
        "nav",
        nav(
            ul(
                For(
                    "item in menu_items",
                    li(a(Var("item.name"), href=Var("item.link"))),
                )
            )
        ),
    )
    menu_template = menu.render(pretty=True)
    print(menu_template)
    # menu = None
    try:
        menu = render(menu_template,
                      menu_items=[
                          menu_url("Home", "home.html"),
                          menu_url("About", "about.html"),
                          menu_url("Contact Us", "contact_us.html")
                      ]
                      )
        return menu
    except (Exception,) as e:
        menu = e
    if isinstance(menu, Exception):
        return False
    print(menu)


if __name__ == '__main__':
    print(block_test())



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing as T
from dataclasses import dataclass

from uidom.dom.elements.buttons import SubmitButton
from uidom.dom.htmlelement import HTMLElement
from valio import StringValidator, Validator

__all__ = [
    # boolean html
    "BooleanLabel",
    "BooleanLegend",
    "BooleanInput",
    "CheckboxInput",
    "RadioInput",
    "BooleanField"
]


@dataclass(eq=False)
class BooleanLabel(HTMLElement):
    label: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(BooleanLabel, self).__init__(self.label, *args, **kwargs)

    def __render__(self, label: str, *args, **kwargs):  # noqa
        return self.html_tags.label(label, *args, **kwargs)


class BooleanLabelValidator(Validator):
    annotation = T.Union[BooleanLabel, None]


@dataclass(eq=False)
class BooleanLegend(HTMLElement):
    label: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(BooleanLegend, self).__init__(self.label, *args, **kwargs)

    def __render__(self, label: str, *args, **kwargs):
        return self.html_tags.legend(label, *args, **kwargs)


@dataclass(eq=False)
class BooleanInput(HTMLElement):
    name: str = StringValidator(logger=False, debug=True)
    type: str = StringValidator(in_choice=["checkbox", "radio"], logger=False, debug=True, default="checkbox")
    checked: T.Union[str, bool] = Validator(in_choice=["unchecked", "checked", "null", True, False],
                                                 logger=False, debug=True, default="checked")

    def __post_init__(self, *args, **kwargs):
        super(BooleanInput, self).__init__(*args,
                                           name=self.name,
                                           type=self.type,
                                           checked=self.checked,
                                           **kwargs)

    def __render__(
            self,
            *args,
            name,
            type,
            checked,
            **kwargs
    ):

        return self.html_tags.input_(
            *args,
            name=name,
            type=type,
            checked=checked if checked != "null" else False,
            **kwargs,
        )


class BooleanInputValidator(Validator):
    annotation = T.Union[BooleanInput, None]


@dataclass(eq=False)
class CheckboxInput(BooleanInput):
    name: str = StringValidator(logger=False, debug=True)
    checked: T.Union[str, bool] = Validator(in_choice=["unchecked", "checked", "null", True, False],
                                            logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(BooleanInput, self).__init__(*args,
                                           name=self.name,
                                           checked=self.checked,
                                           **kwargs)

    def __render__(
            self,
            *args,
            name,
            checked,
            **kwargs
    ):

        return super().__render__(
            *args,
            name=name,
            type="checkbox",
            checked=checked,
            **kwargs,
        )


@dataclass(eq=False)
class RadioInput(BooleanInput):
    name: str = StringValidator(logger=False, debug=True)
    checked: T.Union[str, bool] = Validator(in_choice=["unchecked", "checked", "null", True, False],
                                            logger=False, debug=True, default="checked")

    def __post_init__(self, *args, **kwargs):
        super(BooleanInput, self).__init__(*args,
                                           name=self.name,
                                           checked=self.checked,
                                           **kwargs)

    def __render__(
            self,
            *args,
            name,
            checked,
            **kwargs
    ):
        return super().__render__(
            *args,
            name=name,
            type="radio",
            checked=checked,
            **kwargs,
        )


@dataclass(eq=False)
class BooleanField(HTMLElement):
    labeled = BooleanLabelValidator(logger=False, debug=True)
    input = BooleanInputValidator(logger=False, debug=True)
    label: str
    name: str
    type: str
    checked: T.Union[str, bool]

    def __post_init__(self, *args, **kwargs):
        super(BooleanField, self).__init__(*args,
                                           label=self.label,
                                           name=self.name,
                                           type=self.type,
                                           checked=self.checked,
                                           **kwargs
                                           )

    def __render__(self, *args, label, name, type, checked, **kwargs):
        self.labeled = BooleanLabel(label)
        self.input = BooleanInput(
                  name=name,
                  type=type,
                  checked=checked,
              )
        return self.html_tags.div(*args, self.labeled, self.input, **kwargs)


if __name__ == '__main__':
    from uidom.dom import form
    field = BooleanField(
        label="Terms and Conditions",
        checked="null",
        name="select",
        type="radio")
    button = SubmitButton(label="Agree", value="agree")
    field.labeled["for"] = field.input["id"] = "select"
    button["for"] = field["id"] = "xyz"
    print(form(field, button, method="POST"))
    print(RadioInput(checked="unchecked", name="deliver"))




# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from .direction import *
from .flexes import *
from .grow import *
from .order import *
from .shrink import *
from .wrap import *



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uidom',
 'uidom.asgi_plugs',
 'uidom.components',
 'uidom.components.alpinejs',
 'uidom.components.atoms',
 'uidom.dom',
 'uidom.dom.elements',
 'uidom.dom.iconify',
 'uidom.dom.iconify.account',
 'uidom.dom.iconify.brands',
 'uidom.dom.src',
 'uidom.dom.src.tests',
 'uidom.dom.src.utils',
 'uidom.dom.tailwindcss',
 'uidom.dom.tailwindcss.background',
 'uidom.dom.tailwindcss.flexbox',
 'uidom.dom.tailwindcss.gridbox',
 'uidom.dom.tailwindcss.layout',
 'uidom.dom.tailwindcss.sizing',
 'uidom.dom.tailwindcss.spacing',
 'uidom.dom.tailwindcss.typography',
 'uidom.dom.utils',
 'uidom.dom.widgets',
 'uidom.edge_db',
 'uidom.response',
 'uidom.settings',
 'uidom.web_io']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.1,<4.0.0',
 'aiofiles>=0.7.0,<0.8.0',
 'pytailwindcss>=0.1.4,<0.2.0',
 'python-multipart>=0.0.5,<0.0.6',
 'starlette==0.19.1',
 'valio>=0.1.0-beta.5,<0.2.0']

setup_kwargs = {
    'name': 'uidom',
    'version': '0.1.1b6',
    'description': 'HTML library',
    'long_description': None,
    'author': 'bitplorer',
    'author_email': 'bitplorer@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class Order(HTMLElement):

    def __render__(self, elem, order="1"):
        elem["class"] += f" order-{order}"
        return elem





# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class ClearLeft(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "clear-left"
        return elem


class ClearRight(HTMLElement):

    def __render__(self, elem):
        elem["class"] = "clear-right"
        return elem


class ClearBoth(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "clear-both"
        return elem


class ClearNone(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "clear-none"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import collections
import typing

__all__ = ["map_recursive", "apply_to_list", "apply_to_dict", "apply_to_tuple"]

iter_types = typing.Iterable, typing.Iterator, typing.Mapping


def f(value, key=None): return value if key is None else key, value


def apply_to_list(elem_func, iterable, obj_func=None):
    obj_func = obj_func or f
    return list(
        apply_to_frozenset(elem_func, item, obj_func)
        if isinstance(item, frozenset)
        else apply_to_set(elem_func, item, obj_func)
        if isinstance(item, set)
        else apply_to_list(elem_func, item, obj_func)
        if isinstance(item, list)
        else apply_to_tuple(elem_func, item, obj_func)
        if isinstance(item, tuple)
        else apply_to_dict(elem_func, item, obj_func)
        if isinstance(item, dict)
        else elem_func(item)
        for item in iterable
    )


def apply_to_tuple(elem_func, iterable, obj_func=None):
    obj_func = obj_func or f
    return tuple(
        apply_to_frozenset(elem_func, item, obj_func)
        if isinstance(item, frozenset)
        else apply_to_set(elem_func, item, obj_func)
        if isinstance(item, set)
        else apply_to_tuple(elem_func, item, obj_func)
        if isinstance(item, tuple)
        else apply_to_list(elem_func, item, obj_func)
        if isinstance(item, list)
        else apply_to_dict(elem_func, item, obj_func)
        if isinstance(item, dict)
        else elem_func(item)
        for item in iterable
    )


def apply_to_dict(elem_func, iterable, obj_func=None):
    obj_func = obj_func or f
    return dict(
        obj_func(apply_to_frozenset(elem_func, item, obj_func), key)
        if isinstance(item, frozenset)
        else obj_func(apply_to_set(elem_func, item, obj_func), key)
        if isinstance(item, set)
        else obj_func(apply_to_tuple(elem_func, item, obj_func), key)
        if isinstance(item, tuple)
        else obj_func(apply_to_list(elem_func, item, obj_func), key)
        if isinstance(item, list)
        else obj_func(apply_to_dict(elem_func, item, obj_func), key)
        if isinstance(item, dict)
        else obj_func(elem_func(item), key)
        for key, item in iterable.items()
    )


def apply_to_set(elem_func, iterable, obj_func=None):
    obj_func = obj_func or f
    return set(
        apply_to_frozenset(elem_func, item, obj_func)
        if isinstance(item, frozenset)
        else apply_to_set(elem_func, item, obj_func)
        if isinstance(item, set)
        else apply_to_tuple(elem_func, item, obj_func)
        if isinstance(item, tuple)
        else apply_to_list(elem_func, item, obj_func)
        if isinstance(item, list)
        else apply_to_dict(elem_func, item, obj_func)
        if isinstance(item, dict)
        else elem_func(item)
        for item in iterable
    )


def apply_to_frozenset(elem_func, iterable, obj_func=None):
    obj_func = obj_func or f
    return frozenset(
        apply_to_frozenset(elem_func, item, obj_func)
        if isinstance(item, frozenset)
        else apply_to_set(elem_func, item, obj_func)
        if isinstance(item, set)
        else apply_to_tuple(elem_func, item, obj_func)
        if isinstance(item, tuple)
        else apply_to_list(elem_func, item, obj_func)
        if isinstance(item, list)
        else apply_to_dict(elem_func, item, obj_func)
        if isinstance(item, dict)
        else elem_func(item)
        for item in iterable
    )


def map_recursive(elem_func, iterable, obj_func=None):
    obj_func = obj_func or f
    if isinstance(iterable, dict):
        return apply_to_dict(elem_func, iterable, obj_func)
    if isinstance(iterable, list):
        return apply_to_list(elem_func, iterable, obj_func)
    if isinstance(iterable, tuple):
        return apply_to_tuple(elem_func, iterable, obj_func)
    if isinstance(iterable, set):
        return apply_to_set(elem_func, iterable, obj_func)
    if isinstance(iterable, frozenset):
        return apply_to_frozenset(elem_func, iterable, obj_func)
    if isinstance(iterable, collections.Generator):
        return (map_recursive(elem_func, item, obj_func) for item in iterable)


def map_r(func, iterables):
    return (map_recursive(func, iterable) for iterable in iterables)


def zip_recursive(iterables):
    # TODO make this function
    pass

#
# def _test_dict_map():
#     return map_recursive(
#         lambda x: x if isinstance(x, int) else None,
#         {"prices": [1, [2, [3.0], (9.0, 5)]]},
#     )
#
#
# def _test_generator():
#     return map_recursive(
#         lambda x: x if isinstance(x, int) else None,
#         ({"prices": [1, [2, [3.0], (9.0, 5)]]} for _ in range(3)),
#     )
#
#
# def _test_isinstance(d, f2=None):
#     return map_recursive(lambda x: x if x > 6 else None, d, f2)
#
#
# data1 = (
#     1,
#     {"prices": [(2.0, 3.0), (4.6, 5.6)], "weight": [9.0], "model": [5]},
#     {"prices": [2.0, 3.0], "weight": [9.0, 10.0], "model": [5]},
# )
#
# _ = lambda x: x ** 2
# __ = lambda v, k=None: (k.upper(), v) if k is not None else v
#
# if __name__ == "__main__":
#     data = [[2, 4, 8]]
#     print(map_recursive(_, *data))
#     print(_test_isinstance(data1, __))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement

__all__ = [
    "FlexColumn",
    "FlexRow"
]


class FlexColumn(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-col"
        return elem


class FlexRow(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-row"
        return elem


class FlexColumnReversed(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-col-reverse"
        return elem


class FlexRowReversed(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-row-reverse"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class FlexGrow(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-grow"
        return elem


class FlexGrow0(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-grow-0"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom import *


class Accordian(XComponent):

    def __render__(self, tag_name):
        return template(
            raw('''
    <div x-data="{
        ...accordianParam(),
        ...$el.parentElement.data(),
        get expanded() {
            return this.active === this.id
        },
        set expanded(value) {
            this.active = value ? this.id : null
        }
    }" role="region" class="border border-black p-1 border-collapse">
        <h2>
            <button
                @click="expanded = !expanded"
                :aria-expanded="expanded"
                class="flex items-center justify-between w-full font-bold text-xl px-6 py-3"
                :class="{'': this.active === this.id }"
                x-effect="console.log(this.active)"
            >
                <span x-text="title"></span>
                <span x-show="expanded" aria-hidden="true" class="ml-4">&minus;</span>
                <span x-show="!expanded" aria-hidden="true" class="ml-4">&plus;</span>
            </button>
        </h2>

        <div x-show="expanded" x-collapse>
            <div class="pb-4 px-6" x-html="content"></div>
        </div>
    </div>
            '''),
            script("function accordianParam(){ return {title:'', content:'', id: '', active: '' } }"),
            x_component=tag_name
        )


accordian = Accordian(tag_name="accordian")


class AccordianGroup(XComponent):
    def __render__(self, tag_name):
        return template(div(slot()), x_data="{}", x_component=tag_name)


accordianGroup = AccordianGroup(tag_name="accordian-group")



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class FloatRight(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "float-right"
        return elem


class FloatLeft(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "float-left"
        return elem


class FloatNone(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "float-none"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

'''
This module consists of classes specific to HTML5-SVG Elements. In general this module does not include
- Elements that are not specific to SVG (eg. <a>)
- Elements that are deprecated
'''
from uidom.dom.src.main import extension

# Tag attributes
_ATTR_GLOBAL = set([
    'accesskey', 'class', 'class', 'contenteditable', 'contextmenu', 'dir',
    'draggable', 'id', 'item', 'hidden', 'lang', 'itemprop', 'spellcheck',
    'style', 'subject', 'tabindex', 'title'
])

# https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/Events#Attributes
_ATTR_EVENTS = set([
    'onbegin', 'onend', 'onrepeat',
    'onabort', 'onerror', 'onresize', 'onscroll', 'onunload',
    'oncopy', 'oncut', 'onpaste',
    'oncancel', 'oncanplay', 'oncanplaythrough', 'onchange', 'onclick', 'onclose', 'oncuechange', 'ondblclick',
    'ondrag', 'ondragend', 'ondragenter', 'ondragexit', 'ondragleave', 'ondragover', 'ondragstart', 'ondrop',
    'ondurationchange', 'onemptied', 'onended', 'onerror', 'onfocus', 'oninput', 'oninvalid', 'onkeydown', 'onkeypress',
    'onkeyup', 'onload', 'onloadeddata', 'onloadedmetadata', 'onloadstart', 'onmousedown', 'onmouseenter',
    'onmouseleave', 'onmousemove', 'onmouseout', 'onmouseover', 'onmouseup', 'onmousewheel', 'onpause', 'onplay',
    'onplaying', 'onprogress', 'onratechange', 'onreset', 'onresize', 'onscroll', 'onseeked', 'onseeking', 'onselect',
    'onshow', 'onstalled', 'onsubmit', 'onsuspend', 'ontimeupdate', 'ontoggle', 'onvolumechange', 'onwaiting'
])

DASHED_ATTRIBUTES = set([
    'accent', 'alignment', 'arabic', 'baseline', 'cap', 'clip', 'color', 'dominant', 'enable', 'fill', 'flood',
    'font', 'glyph', 'horiz', 'image', 'letter', 'lighting', 'marker', 'overline', 'paint', 'panose', 'pointer',
    'rendering', 'shape', 'stop', 'strikethrough', 'stroke', 'text', 'underline', 'unicode', 'units', 'v', 'vector',
    'vert', 'word', 'writing', 'x'
])


# https://developer.mozilla.org/en-US/docs/Web/SVG/Element/svg
class svg_tag(extension.Tags):
    @staticmethod
    def clean_attribute(attribute):
        attribute = extension.Tags.clean_attribute(attribute)
        words = attribute.split('_')
        if words[0] in DASHED_ATTRIBUTES:
            return attribute.replace('_', '-')
        return attribute


class svg(svg_tag):
    pass


class animate(svg_tag):
    '''
    The  animate SVG element is used to animate an attribute or property of an element over time.
    It's normally inserted inside the element or referenced by the href attribute of the target element.
    '''
    pass


class animateMotion(svg_tag):
    '''
    The <animateMotion> element causes a referenced element to move along a motion path.
    '''
    pass


class animateTransform(svg_tag):
    '''
    The animateTransform element animates a transformation attribute on its target element, thereby allowing
    animations to control translation, scaling, rotation, and/or skewing.
    '''
    is_single = True


class circle(svg_tag):
    '''
    The <circle> SVG element is an SVG basic shape, used to draw circles based on a center point and a radius.
    '''
    pass


class clipPath(svg_tag):
    '''
    The <clipPath> SVG element defines a clipping path, to be used used by the clip-path property.
    '''
    pass


class defs(svg_tag):
    '''
    The <defs> element is used to store graphical objects that will be used at a later time. Objects created inside a
    <defs> element are not rendered directly. To display them you have to reference them
    (with a <use> element for example).
    '''
    pass


class desc(svg_tag):
    '''
    The <desc> element provides an accessible, long-text description of any SVG container element or graphics element.
    '''
    pass


class ellipse(svg_tag):
    '''
    An ellipse element for svg containers
    '''
    pass


# (Note, filters are at the bottom of this file)

class g(svg_tag):
    '''
    The <g> SVG element is a container used to group other SVG elements.
    '''
    pass


class image(svg_tag):
    '''
    The <image> SVG element includes images inside SVG documents. It can display raster image files or other SVG files.
    '''
    pass


class line(svg_tag):
    '''
    The <line> element is an SVG basic shape used to create a line connecting two points.
    '''
    pass


class linearGradient(svg_tag):
    '''
    The <linearGradient> element lets authors define linear gradients that can be applied to fill or
    stroke of graphical elements.
    '''
    pass


class marker(svg_tag):
    '''
    The <marker> element defines the graphic that is to be used for drawing arrowheads or polymarkers on a given <path>, <line>, <polyline> or <polygon> element.
    '''
    pass


class mask(svg_tag):
    '''
    The <mask> element defines an alpha mask for compositing the current object into the background.
    A mask is used/referenced using the mask property.
    '''
    pass


class mpath(svg_tag):
    '''
    The <mpath> sub-element for the <animateMotion> element provides the ability to reference an
    external <path> element as the definition of a motion path.
    '''
    pass


class pattern(svg_tag):
    '''
    The <pattern> element defines a graphics object which can be redrawn at repeated x and y-coordinate
    intervals ("tiled") to cover an area.
    '''
    pass


class polygon(svg_tag):
    '''
    A polygon element for svg containers
    '''
    pass


class polyline(svg_tag):
    '''
    A polyline element for svg containers
    '''
    pass


class radialGradient(svg_tag):
    '''
    The <radialGradient> element lets authors define radial gradients that can be applied to fill
    or stroke of graphical elements.
    '''
    pass



class path(svg_tag):
    '''
    A path element for svg containers
    '''
    pass


class rect(svg_tag):
    '''
    A rectangle element for svg containers
    '''
    pass


class stop(svg_tag):
    '''
    The SVG <stop> element defines a color and its position to use on a gradient.
    This element is always a child of a <linearGradient> or <radialGradient> element.
    '''
    pass


class switch(svg_tag):
    '''
    The <switch> SVG element evaluates any requiredFeatures, requiredExtensions and systemLanguage attributes
    on its direct child elements in order, and then renders the first child where these attributes evaluate to true.
    Other direct children will be bypassed and therefore not rendered. If a child element is a container element,
    like <g>, then its subtree is also processed/rendered or bypassed/not rendered.
    '''
    pass


class symbol(svg_tag):
    '''
    The use of symbol elements for graphics that are used multiple times in the same document adds structure and
     semantics. Documents that are rich in structure may be rendered graphically, as speech, or as Braille,
     and thus promote accessibility.
    '''
    pass


class text(svg_tag):
    '''
    The SVG <text> element draws a graphics element consisting of text. It's possible to apply a gradient,
     pattern, clipping path, mask, or filter to <text>, like any other SVG graphics element.
    '''
    pass


class textPath(svg_tag):
    '''
    To render text along the shape of a <path>, enclose the text in a <textPath> element that has an href
    attribute with a reference to the <path> element.
    '''
    pass


class title(svg_tag):
    '''
    The <title> element provides an accessible, short-text description of any SVG container
    element or graphics element.
    '''
    pass


class tspan(svg_tag):
    '''
    The SVG <tspan> element define a subtext within a <text> element or another <tspan> element.
    It allows to adjust the style and/or position of that subtext as needed.
    '''
    pass


class use(svg_tag):
    '''
    The <use> element takes nodes from within the SVG document, and duplicates them somewhere else.
    '''
    pass


class view(svg_tag):
    '''
    A view is a defined way to view the image, like a zoom level or a detail view.
    '''
    pass


# FILTERS
class filter(svg_tag):
    pass


class feBlend(svg_tag):
    pass


class feColorMatrix(svg_tag):
    pass


class feComponentTransfer(svg_tag):
    pass


class feComposite(svg_tag):
    pass


class feConvolveMatrix(svg_tag):
    pass


class feDiffuseLighting(svg_tag):
    pass


class feDisplacementMap(svg_tag):
    pass


class feFlood(svg_tag):
    pass


class feGaussianBlur(svg_tag):
    pass


class feImage(svg_tag):
    pass


class feMerge(svg_tag):
    pass


class feMorphology(svg_tag):
    pass


class feOffset(svg_tag):
    pass


class feSpecularLighting(svg_tag):
    pass


class feTile(svg_tag):
    pass


class feTurbulence(svg_tag):
    pass


class feDistantLight(svg_tag):
    pass


class fePointLight(svg_tag):
    pass


class feSpotLight(svg_tag):
    pass



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class Container(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " container"
        return elem


class Small(HTMLElement):

    def __render__(self, elem):
        elem["class"] = ' '.join(map(lambda cls: f" sm:{cls}"
        if any(cls) and ":" not in cls else "" if not any(cls) else cls, elem["class"].split(" ")))
        return elem


class Medium(HTMLElement):

    def __render__(self, elem):
        elem["class"] = ' '.join(map(lambda cls: f" md:{cls}"
        if any(cls) and ":" not in cls else "" if not any(cls) else cls, elem["class"].split(" ")))
        return elem


class Large(HTMLElement):

    def __render__(self, elem):
        elem["class"] = ' '.join(map(lambda cls: f" lg:{cls}"
        if any(cls) and ":" not in cls else "" if not any(cls) else cls, elem["class"].split(" ")))
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing as T
from dataclasses import dataclass

from uidom.dom.htmlelement import HTMLElement
from uidom.dom.src.htmltags import html_tag
from valio import StringValidator, Validator

__all__ = [
    "Button",
    "SubmitButton",
    "ResetButton",
    "ButtonInput",
    "SubmitButtonInput",
    "ResetButtonInput",
    "ImageButtonInput",
    "FileButtonInput",
]


@dataclass(eq=False)
class _Button(HTMLElement):

    label: str = StringValidator(logger=False, debug=True)
    type: str = StringValidator(in_choice=["submit", "reset", "button"], logger=False, debug=True)
    value: str = StringValidator(logger=False, debug=True)
    icon: T.Union[html_tag, None] = Validator(logger=False, debug=True, default=None)

    def __post_init__(self, *args, **kwargs):
        super(_Button, self).__init__(
            *args,
            label=self.label,
            type=self.type,
            value=self.value,
            icon=self.icon if self.icon is not None else False,
            **kwargs)

    def __render__(self, *args, label, type, value, icon=None, **kwargs):
        return self.html_tags.button(icon or '', label, *args, type=type, value=value, **kwargs)


@dataclass(eq=False)
class _ButtonInput(HTMLElement):
    type: str = StringValidator(in_choice=["submit", "reset", "button", "file", "image"], logger=False, debug=True)
    value: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(_ButtonInput, self).__init__(*args, type=self.type, value=self.value, **kwargs)

    def __render__(self, *args, type, value, **kwargs):
        return self.html_tags.input_(*args, type=type, value=value, **kwargs)


@dataclass(eq=False)
class Button(HTMLElement):
    label: str
    value: str

    def __post_init__(self, *args, **kwargs):
        super(Button, self).__init__(*args, label=self.label, value=self.value, **kwargs)

    def __render__(self, *args, label, value, **kwargs):
        return _Button(label=label, type="button", value=value)


@dataclass(eq=False)
class SubmitButton(HTMLElement):
    label: str
    value: str
    icon: T.Union[html_tag, None] = None

    def __post_init__(self, *args, **kwargs):
        super(SubmitButton, self).__init__(*args, label=self.label, value=self.value, icon=self.icon, **kwargs)

    def __render__(self, *args, value, label, icon, **kwargs):
        return _Button(type="submit", value=value, label=label, icon=icon)


@dataclass(eq=False)
class ResetButton(HTMLElement):
    label: str
    value: str
    icon: T.Union[html_tag, None] = None

    def __post_init__(self, *args, **kwargs):
        super(ResetButton, self).__init__(*args, label=self.label, value=self.value, icon=self.icon, **kwargs)

    def __render__(self, *args, label, value, icon, **kwargs):
        return _Button(*args, type="reset", label=label, value=value, icon=icon, **kwargs)

@dataclass(eq=False)
class ButtonInput(HTMLElement):
    value: str
    
    def __render__(self, *args, value, **kwargs):
        return _ButtonInput(type="button", value=value)

@dataclass(eq=False)
class SubmitButtonInput(HTMLElement):
    value: str 
    
    def __render__(self, *args, value, **kwargs):
        return _ButtonInput(type="submit", value=value)

@dataclass(eq=False)
class ResetButtonInput(HTMLElement):
    value: str
    
    def __render__(self, *args, value, **kwargs):
        return _ButtonInput(type="reset", value=value)

@dataclass(eq=False)
class FileButtonInput(HTMLElement):
    value: str 
    
    def __render__(self, *args, value, **kwargs):
        return _ButtonInput(type="file", value=value)

@dataclass(eq=False)
class ImageButtonInput(HTMLElement):
    value: str 
    
    def __render__(self, *args, value, **kwargs):
        return _ButtonInput(type="image", value=value)


if __name__ == '__main__':
    print(SubmitButton(label="submit", value="upload"))
    print(ResetButton(label="reset", value="reset"))
    # print(FileButtonInput(value="a"))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from .booleans import *
from .buttons import *
from .chars import *
from .dates import *
from .enums import *
from .floats import *
from .integers import *



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.src.main import extension

__all__ = [
    "CSSClass",
    "CSSId",
    "CSS",
    "CSSProperty"
]


class CSS(extension.StyleTags):
    pass


class CSSClass(extension.StyleTags):
    is_class = True
    is_inline = True


class CSSId(extension.StyleTags):
    is_id = True
    is_inline = True


class CSSProperty(extension.StyleTags):

    def __set_name__(self, owner, name):
        self._id = f"{owner.__name__}.{name}".replace(".", "-").replace("_", "-")
        self._class = name.replace("_", "-")
        self._apply = f"{owner.__name__}::part({name})".replace("_", "-")

        class Class(CSSClass):
            tagname = self._class

        class Id(CSSId):
            tagname = self._id.lower()

        class CSS(extension.StyleTags):
            tagname = name.replace("_", "-")

        class Apply(extension.StyleTags):
            tagname = self._apply.lower()
            is_apply = True
            left_delimiter = "{@"

        self.css_class = Class
        self.css_id = Id
        self.css = CSS
        self.css_apply = Apply


if __name__ == '__main__':
    # from myst_parser import parse_html

    class Order(object):
        product_name = CSSProperty()
        product_category = CSSProperty()
        product_id = CSSProperty()


    order = Order()
    products = order.product_name.css_id(background_color='red', width="12%")
    category = order.product_category.css_class(color="#23cfff")
    csss = order.product_id.css_apply(apply="bg-rose-500")
    id = order.product_id.css_id()


    # ast = parse_html.tokenize_html(html(style(products, category)).render())
    # print(list(ast.walk()))

    class product(CSSClass):
        pass


    cs = products
    # cs.attributes.update(padding="10px")
    cs["boxsize"] = "border-box"
    cs["background-color"] = "red"
    print(cs)
    print(category)
    print(csss.render(pretty=False))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom import HTMLElement, raw, script

__all__ = [
    "VueLoader",
]


class VueLoader(HTMLElement):
    # for more details look at
    # https://github.com/FranckFreiburger/vue3-sfc-loader

    def __render__(self, *args, **kwargs):
        return script(
            raw('''
        // Options from the vue3-sfc-loader Example at https://github.com/FranckFreiburger/vue3-sfc-loader#example
        const options = {
            moduleCache: {
                vue: Vue
            },
            async getFile(url) {
                const res = await fetch(url);
                if (!res.ok)
                    throw Object.assign(new Error(res.statusText + ' ' + url), { res });
                return {
                    getContentData: asBinary => asBinary ? res.arrayBuffer() : res.text(),
                }
            },
            addStyle(textContent) {
                const style = Object.assign(document.createElement('style'), { textContent });
                const ref = document.head.getElementsByTagName('style')[0] || null;
                document.head.insertBefore(style, ref);
            },
        }
        const { loadModule } = window['vue3-sfc-loader'];
        const app = createApp({
            components: {
                'vue-card': vue_card, // Vue Component from Razor Partial
                'vue-card2': Vue.defineAsyncComponent(() => loadModule('./js/vue-card2.vue', options)) // Vue SFC
            }
        });
        app.mount('#app');
        '''),
            type="text/javascript"
        )
